#!/usr/bin/perl -w

##############################################################################
# The purpose of this script is to summarize the results contained in the
# output file of mbc-evalCandGenesByCoExpr-leaveOneOut.pl
#
# copyright 2007 by Rosario M. Piro (rosario.piro@unito.it)
#
# ---------------------------------------------------------------------------
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; _exclusively_ version 2 of the License
#    NO later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
##############################################################################

use strict;

use Getopt::Std;
use File::Basename;
use POSIX qw(strtod);

# command buffering of STDOUT
$| = 1;

my $prgname = "mbc-evalCandGenesByCoExpr-leaveOneOut-extractDisGenes.pl";

my $resultsFile = "";

my $extendedOutput = 0;
my $verboseOutput = 0;



# -------------------
# getting options ...
# -------------------

my %opts;
$opts{h} = "";
$opts{r} = "";
$opts{e} = "";
$opts{v} = "";

getopts('hr:ev', \%opts);

if ( $opts{h} eq "1" ) {
    &printUsageInfo();
}

if ( $opts{r} ne "" ) {
    $resultsFile = $opts{r};
}

if ( $opts{e} eq "1" ) {
    $extendedOutput = 1;
}

if ( $opts{v} eq "1" ) {
    $verboseOutput = 1;
}

# ---------------
# checking files:
# ---------------

if ( $resultsFile eq "") {
    print STDERR "Error: At least one of the mandatory parameters is missing! (-h for help)\n";
    exit(1);
}

if (! -e $resultsFile || ! -r $resultsFile) {
    print STDERR "Error: $resultsFile does not exist or cannot be read!\n";
    exit(2);
}


# ----------------------------
# read and process input file:
# ----------------------------

unless ( open(RESIN, "< $resultsFile") ) {
    print STDERR "Error: cannot open input file $resultsFile!\n";
    exit(1);
}

if ($verboseOutput) {
    print "# Reading input file ".basename($resultsFile)." ...\n";
}

my $line = "";
my $lineNumber = 0;

my $thisPhenotype = "";
my $thisDisGene = "";

my $header = "";
if ($extendedOutput) {
    $header .= "OMIM_id\t";
    $header .= "refGenes\t";
    $header .= "avPairwRefCorrScore\t";
    $header .= "refG4OMIM_id\t";
    $header .= "candGenes\t";
}

my $headerWritten = 0;

my $numRefGenes4ThisPhenotype = 0;
my $numRefGenes = 0;
my $numCandGenes = 0;
my $avRefCorrScore = "";

while ($line = <RESIN>) {

    $lineNumber++;

    chomp($line);

    if ($verboseOutput) {
	print "# parsing line: $line\n";
    }

    if ($line =~ /^\+\+\+ OMIM ID=(\d+)[^\d].* \+\+\+$/) {
	# new phenotype!
	$thisPhenotype = $1;
    }
    elsif ($line =~ /^--- OMIM ID=$thisPhenotype gene=(.+) ---$/) {
	# new phenotype-disease gene pair
	$thisDisGene = $1;
    }
    elsif ($line =~ /^\# Number of reference genes with expression profile associated to OMIM ID \d+: ([\d\.]+)/) {
	$numRefGenes4ThisPhenotype = $1;
    }
    elsif ($line =~ /^\# Number of reference genes having expression data: ([\d\.]+)/) {
	$numRefGenes = $1;
    }
    elsif ($line =~ /^\# Average pairwise correlation score of reference genes: ([^ ]+)/) {
	$avRefCorrScore = $1;
    }
    elsif ($line =~ /^\# Number of candidate genes having necessary data for scoring[^:]*: (\d+)/) {
	$numCandGenes = $1;
    }
    elsif (!$headerWritten && $line =~ /^rank\s/) { # the original header
	# write header:
	print "$header$line\n"; # add exentual extended header at beginning!
	$headerWritten = 1;
    }
    elsif ($line =~ /^\d+\s$thisDisGene\s/) {
	# this is the entry of the corresponding disease gene! (first column: rank; second column: geneID
	if ($extendedOutput) {
	    print "$thisPhenotype\t";
	    print "$numRefGenes\t";
	    print "$avRefCorrScore\t";
	    print "$numRefGenes4ThisPhenotype\t";
	    print "$numCandGenes\t";
	}
	print "$line\n";
    }
}

close(RESIN);

exit(0);

#############################################################################
#                                   FUNCTIONS:
#############################################################################


# ---------------------------------------------------------------------------
#
# Print usage information (help).
# arguments: none
#
# --------------------------------------------------------------------------
sub printUsageInfo() {
    print "Author: Rosario M. Piro (piro\@to.infn.it)\n\n";
    print "Usage:\n";
    print "$prgname [OPTIONS]\n";
    print "\nWhere OPTIONS are:\n";
    print "-h                Display this help message and exit.\n";
    print "-r <infile>       File containing the prioritization results (output of\n";
    print "                  mbc-evalCandGenesByCoExpr-leaveOneOut.pl).\n";
    print "-e                Extended output: add OMIM phenotype ID to each line,\n";
    print "                  as well as the number of candidates and reference genes\n";
    print "                  that have expression profiles, as well as the number of\n";
    print "                  reference genes (with expression profiles) directly involved\n";
    print "                  in the query disease.\n";
    print "-v                Verbose stdout.\n";
    print "Mandatory arguments are: -r.\n\n";

    exit(0);

} # printUsageInfo() ...
