#!/usr/bin/perl -w

##############################################################################
# The purpose of this script is to merge multiple results contained from
# mbc-evalCandGenesByCoExpr-leaveOneOut.pl
#
# copyright 2007-2011 by Rosario M. Piro (rosario.piro@unito.it;r.piro@dkfz.de)
#
# ---------------------------------------------------------------------------
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; _exclusively_ version 2 of the License
#    NO later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
##############################################################################

use strict;

use Getopt::Std;
use File::Basename;
use POSIX qw(strtod);

# definitions

my $NUM_REF_4_PHENO = "Number of reference genes with expression profile associated to OMIM ID";
my $NUM_REF_EXPR =    "Number of reference genes having expression data";
my $AV_PW_CORR =      "Average pairwise correlation score of reference genes";
my $NUM_CAN_SCORED =  "Number of candidate genes having necessary data for scoring";

# command buffering of STDOUT
$| = 1;

my $prgname = "mbc-evalCandGenesByCoExpr-leaveOneOut-mergeLogFilesByScore.pl";

my $resultFiles = "";
my $scoreMapFiles = "";
my $scoringFunct = "average";

my $rankProdScoring = 0;

my $strictScoring = 0;

my $extendedOutput = 0;
my $verboseOutput = 0;


# global variables that we need for some subroutines also:

my %gene2score = ();




# -------------------
# getting options ...
# -------------------

my %opts;
$opts{h} = "";
$opts{r} = "";
$opts{M} = "";
$opts{R} = "";
$opts{s} = "";
$opts{S} = "";
$opts{e} = "";
$opts{v} = "";

getopts('hr:M:Rs:Sev', \%opts);

if ( $opts{h} eq "1" ) {
    &printUsageInfo();
}

if ( $opts{r} ne "" ) {
    $resultFiles = $opts{r};
}

if ( $opts{M} ne "" ) {
    $scoreMapFiles = $opts{M};
}

if ( $opts{R} eq "1" ) {
    $rankProdScoring = 1;
}

if ( $opts{s} ne "" ) {
    $scoringFunct = $opts{s};
}

if ( $opts{S} eq "1" ) {
    $strictScoring = 1;
}

if ( $opts{e} eq "1" ) {
    $extendedOutput = 1;
}

if ( $opts{v} eq "1" ) {
    $verboseOutput = 1;
}


# checking parameters:

if ($scoringFunct ne "average"
    && $scoringFunct ne "rankproduct"
    && $scoringFunct ne "maximum"
    && $scoringFunct ne "minimum"
    && $scoringFunct ne "allmaximum"
    && $scoringFunct ne "allminimum"
    && $scoringFunct ne "rank2scoreproduct") {

    print STDERR "Error: unknown scoring function \"$scoringFunct\" (-h for help)!\n";
    exit(1);
}

if ($scoringFunct eq "rank2scoreproduct") {
    $rankProdScoring = 0; # ignore this ...
}


# ---------------
# checking files:
# ---------------

if ( $resultFiles eq "") {
    print STDERR "Error: At least one of the mandatory parameters is missing! (-h for help)\n";
    exit(1);
}

my @rFiles = split(/,/, $resultFiles);

foreach my $rF (@rFiles) {
    if (! -e $rF || ! -r $rF) {
	print STDERR "Error: $rF does not exist or cannot be read!\n";
	exit(2);
    }
}


my @r2sFiles = ();

if ($scoringFunct eq "rank2scoreproduct") {
    if ($scoreMapFiles eq "") {
	print STDERR "Error: The use of -s \"rank2scoreproduct\" requires -M (see -h for help)!\n";
	exit(1);
    }
    else { # read the rank2score maps!
	@r2sFiles = split(/,/, $scoreMapFiles);

	if (scalar(@r2sFiles) != scalar(@rFiles)) {
	    print STDERR "Error: the number of files specified with -r and -M must be equal!\n";
	    exit(115);
	}

	foreach my $r2sF (@r2sFiles) {
	    if (! -e $r2sF || ! -r $r2sF) {
		print STDERR "Error: $r2sF does not exist or cannot be read!\n";
		exit(2);
	    }
	}
    }
}


# -----------------------------
# read and process input files:
# -----------------------------

# first the rank2score maps, in case we need them:
my %r2sMaps = ();

my $fileNum;
for ($fileNum = 0; $fileNum < scalar(@r2sFiles); $fileNum++) {
    my %map = ();

    unless ( open(R2SF, "< $r2sFiles[$fileNum]") ) {
	print STDERR "Error: cannot open input file $r2sFiles[$fileNum]!\n";
	exit(1);
    }

    if ($verboseOutput) {
	print "# reading rank-to-score map file: $r2sFiles[$fileNum]\n";
    }

    my $line;
    my $lineNum = 0;
    while($line = <R2SF>) {
	chomp($line);
	$lineNum++;
	my @toks = split(/\t/, $line);
	if (scalar(@toks) < 2) {
	    print STDERR "Error: less than 2 fields in line $lineNum of file $r2sFiles[$fileNum]!\n";
	    exit(221);
	}

	if ($verboseOutput) {
	    print "#    mapping rank $toks[0] to score: $toks[1]\n";
	}
	$map{$toks[0]} = $toks[1]; # map this rank to this score
    }

    # store the map as associated to the Nth result file!
    $r2sMaps{$rFiles[$fileNum]} = \%map;

    close(R2SF);
}


# now open the result files
my @RFILES = (); # to store file handles!
my %rfhandle2rfid = ();
my %rfid2rfname = ();

my $fid = 0;

foreach my $rF (@rFiles) {

    my $RF;

    # open
    unless ( open($RF, "< $rF") ) {
	print STDERR "Error: cannot open input file $rF!\n";
	exit(1);
    }

    push(@RFILES, $RF);
    $rfhandle2rfid{$RF} = ++$fid;
    $rfid2rfname{$fid} = $rF;

    print "# Result file id=$fid; name: ".basename($rF)."\n";

    if ($verboseOutput) {
	print "# Reading input file ".basename($rF)." ...\n";
    }
}

print "# scoring function for merge: $scoringFunct\n";


my %lines = ();
my $lineNumber = 0;

my $thisPhenotype = "";
my $thisDisGene = "";
my $rankHeader = "rank\tgene\tscore_$scoringFunct\tscore_av\trelRank_scored\trelRank_all";

my %gene2ranks = ();
my @candidateGenes = ();
my @candidateGenesAll = ();

my $numRefGenes4ThisPhenotype = 0;
my $numRefGenes = 0;
my $numCandGenesAvg = 0;
my $avRefCorrScore = "-"; # default: undefined
my $numAvRefCorrScores = 0;

my $necData4Scoring = "";

my $keepGoing = 1;

my @runLines = ();
my @infoLines = ();


while ($keepGoing) {

    %lines = ();

    $lineNumber++;

    foreach my $RF (@RFILES) {
	$lines{$RF} = <$RF>;

	if (!$lines{$RF}) { # nothing more to do ...
	    $keepGoing = 0;
	    last;
	}

	chomp($lines{$RF});

	if ($verboseOutput) {
	    print "# parsing line $lineNumber of file $rfhandle2rfid{$RF}: $lines{$RF}\n";
	}
    }

    # nothing more to do ...
    if (!$keepGoing) {
	last;
    }


    foreach my $RF (@RFILES) {
	if ($lines{$RF} =~ /^\+\+\+ OMIM ID=(\d+) gene=(\d+) \+\+\+$/) {
	    # new phenotype?
	    if ($1 ne $thisPhenotype || $2 ne $thisDisGene) {
		# either another phenotype or disease gene => new run!

		# first compute and write rank scores for previous genes
		#if (scalar(keys(%gene2ranks)) != 0) {
		if ($thisPhenotype ne "" && $thisDisGene ne "") {
		    &writeRunAndInfoLines();
		    &scoreAndWriteGenes();
		}

		# initialize stuff
		%gene2ranks = ();
		@candidateGenes = ();
		@candidateGenesAll = ();
		$numRefGenes4ThisPhenotype = 0;
		$numRefGenes = 0;
		$numCandGenesAvg = 0;
		$avRefCorrScore = "-";
		$numAvRefCorrScores = 0;
		$necData4Scoring = "";
		@infoLines = ();
		@runLines = ();

		$thisPhenotype = $1;
		$thisDisGene = $2;

		push(@runLines, $lines{$RF});
	    }
	}
	elsif ($lines{$RF} =~ /^--- OMIM ID=$thisPhenotype gene=(.+) ---$/) {
	    if (scalar(@runLines) < 2) {
		# still need this line, but only once
		push(@runLines, $lines{$RF});
	    }
	}
	elsif ($lines{$RF} =~ /^\# $NUM_REF_4_PHENO \d+: (\d+)/) {

	    # adding to average number of ref. genes for this OMIM
	    $numRefGenes4ThisPhenotype += ( $1/scalar(@RFILES) );

	}
	elsif ($lines{$RF} =~ /^\# $NUM_REF_EXPR: (\d+)/) {

	    # adding to average number of ref. genes
	    $numRefGenes += ( $1/scalar(@RFILES) );

	}
	elsif ($lines{$RF} =~ /^\# $AV_PW_CORR: ([^ ]+)/) {

	    # adding to average of averages
	    if ($1 ne "-") {
		if ($avRefCorrScore eq "-") {
		    $avRefCorrScore = 0;
		}
		$avRefCorrScore += $1;
		$numAvRefCorrScores++;
	    }

	}
	elsif ($lines{$RF} =~ /^\# $NUM_CAN_SCORED([^:]*): (\d+)/) {

	    my $thisNCand = $2;

	    # adding to average number of candidate genes
	    $numCandGenesAvg += ( $thisNCand/scalar(@RFILES) );

	    # keep track what data was necessary for scoring, for all files! 
	    my $nD4S = $1;
	    $nD4S =~ s/^.*\(//;
	    $nD4S =~ s/\).*$//;

	    if ($nD4S ne "") {
		if ($necData4Scoring eq "") {
		    $necData4Scoring = " ($rfhandle2rfid{$RF}=$nD4S)";
		}
		else {
		    $necData4Scoring =~ s/\).*$//;
		    $necData4Scoring .= "|$rfhandle2rfid{$RF}=$nD4S)";
		}
	    }

	}
	elsif ($lines{$RF} =~ /^\#/) {

	    # for all other lines: keep info
	    my $found = 0;
	    my $ii;
	    for ($ii = 0; $ii < scalar(@infoLines); $ii++) {

		$lines{$RF} =~ s/\s*$//; # remove trailing spaces ...
		my $line = $lines{$RF};
		$line =~ s/([\(\)\+\-])/\\$1/g; # escape certain characters

		if ($infoLines[$ii] =~ /^$line \[files: (.*)\]$/) {
		    # we have this already, just update file info:
		    $infoLines[$ii] = "$lines{$RF} [files: $1,$rfhandle2rfid{$RF}]";
		    $found = 1;
		}
	    }
	    if (!$found) {
		# no other similar line; add for the first time:
		push(@infoLines, "$lines{$RF} [files: $rfhandle2rfid{$RF}]");
	    }

	}
	elsif ($lines{$RF} =~ /^rank\s/) { # the original header
	    # do nothing, will be replaced
	}
	elsif ($lines{$RF} =~ /^(\d+)\s([^\s]+)\s([^\s]+)\s/
	       && $lines{$RF} !~ /unscored/) {
	    # this is some gene ranking, store it for later processing!

	    my $found = 0;
	    foreach my $cG (@candidateGenes) {
		if ($cG eq $2) {
		    $found = 1;
		}
	    }
	    if (!$found) {
		push(@candidateGenes, $2);
	    }

	    $found = 0;
	    foreach my $cG (@candidateGenesAll) {
		if ($cG eq $2) {
		    $found = 1;
		}
	    }
	    if (!$found) {
		push(@candidateGenesAll, $2);
	    }

	    if ($rankProdScoring) {
		$gene2ranks{$RF}{$2} = $3;

		if($verboseOutput) {
		    print "# found score of gene $2 for file $RF: $gene2ranks{$RF}{$2}\n"; 
		}
	    }
	    else {
		$gene2ranks{$RF}{$2} = $1;

		if($verboseOutput) {
		    print "# found rank of gene $2 for file $RF: $gene2ranks{$RF}{$2}\n"; 
		}
	    }
	}
	elsif ($lines{$RF} =~ /^(\d+)\s([^\s]+)\s.*unscored/) {
	    # this is some unscored gene, keep at least track of it!

	    my $found = 0;
	    foreach my $cG (@candidateGenesAll) {
		if ($cG eq $2) {
		    $found = 1;
		}
	    }
	    if (!$found) {
		push(@candidateGenesAll, $2);
	    }

	}
    }

}

foreach my $RF (@RFILES) {
    close($RF);
}


# do we still have to write from the last round?
#if (scalar(keys(%gene2ranks)) != 0) {
    &writeRunAndInfoLines();
    &scoreAndWriteGenes();
#}



exit(0);

#############################################################################
#                                   FUNCTIONS:
#############################################################################


# ---------------------------------------------------------------------------
#
# Print usage information (help).
# arguments: none
#
# --------------------------------------------------------------------------
sub printUsageInfo() {
    print "Author: Rosario M. Piro (r.piro\@dkfz.de)\n\n";
    print "Usage:\n";
    print "$prgname [OPTIONS]\n";
    print "\nWhere OPTIONS are:\n";
    print "-h                Display this help message and exit.\n";
    print "-r <infiles>      A comma-separated list of files containing the prioritization\n";
    print "                  results (output of mbc-evalCandGenesByCoExpr-leaveOneOut.pl).\n";
    print "-M <mapfiles>     A comma-separated list of files mapping absolute ranks to scores.\n";
    print "                  This is used only for \"rank2scoreproduct\" (see below).\n";
    print "                  Important: The list must contain the same number of files (in the\n";
    print "                  same order) as -r. Each file must contain a mapping or ranks to\n";
    print "                  scores (one rank per line) in the form: <rank><tab><score>.\n";
    print "-R                Base a gene's overall scoring on its rank product scores\n";
    print "                  (score_mult) of from the single log files instead of its single\n";
    print "                  ranks!\n";
    print "                  Note: This will be ignored for \"rank2scoreproduct\" (see below).\n";
    print "-s <function>     Function to compute overall gene scores. Available functions:\n";
    print "                  \"average\": average rank over <infiles> (default)\n";
    print "                  \"rankproduct\": product of relative ranks in <infiles>\n";
    print "                      The rankproduct of a gene is normalized for the number\n";
    print "                      of <infiles> in which the gene could be scored.\n";
    print "                  \"maximum\": (single) maximum relative rank in <infiles>\n";
    print "                  \"minimum\": (single) minimum relative rank in <infiles>\n";
    print "                  \"allmaximum\": maximum relative ranks in <infiles>\n";
    print "                      If two genes have the same maximum rank, they will be\n";
    print "                      compared by their second worst rank, third worst rank,\n";
    print "                      and so on. This allows for better disambiguation\n";
    print "                      between ties that can easily occur with \"maximum\".\n";
    print "                  \"allminimum\": minimum relative ranks in <infiles>\n";
    print "                      If two genes have the same minimum rank, they will be\n";
    print "                      compared by their second best rank, third best rank,\n";
    print "                      and so on. This allows for better disambiguation\n";
    print "                      between ties that can easily occur with \"minimum\".\n";
    print "                  \"rank2scoreproduct\": the absolute rank of candidate gene is\n";
    print "                      mapped, individually for each result file, to a previously\n";
    print "                      specified score. Then, the obtained scores for the single\n";
    print "                      result files are multiplied to obtain an overall score for\n";
    print "                      the gene. Genes with higher scores are ranked first.\n";
    print "                      This can be used, for example, to implement a scoring\n";
    print "                      scheme that is similar to a noisy-or.\n";
    print "                      For disambiguation, candidates with equal scores will be\n";
    print "                      further prioritized using the procedure described for\n";
    print "                      \"allminimum\" (see above).\n";
    print "                      Note: This requires -M to be scpecified!\n";
    print "                  IMPORTANT: The final ranking will be determined according to\n";
    print "                   _increasing_ overall scores (rank 1 = lowest overall score)\n";
    print "-S                Strict scoring: compute overall scores only for genes\n";
    print "                  that have been scored in all <infiles> (no unscored).\n";
    print "-e                Extended output: add the single ranks obtained by the genes\n";
    print "                  in the different result files.\n";
    print "-v                Verbose stdout.\n";
    print "Mandatory arguments are: -r.\n\n";

    exit(0);

} # printUsageInfo() ...



sub getAndSortRanks() {
    my $gene = $_[0];
    my @ranks = ();
    foreach my $RF (@RFILES) {
	if (exists($gene2ranks{$RF}{$gene})) {
	    push(@ranks, $gene2ranks{$RF}{$gene});
	}
    }

    @ranks = sort {$a <=> $b} @ranks;

    #print "ranks of $gene: @ranks\n";

    return \@ranks;
}


sub cmpAllMinimum() {
    my @ranksA = @{&getAndSortRanks($a)};
    my @ranksB = @{&getAndSortRanks($b)}; 

    #print "comparing: a=$a; ranks: @ranksA\n";
    #print "         : b=$b; ranks: @ranksB\n";

    my $numRanks = scalar(@ranksA);
    if ($numRanks > scalar(@ranksB)) {
	$numRanks = scalar(@ranksB);
    }

    my $ii;
    for ($ii = 0; $ii < $numRanks; $ii++) {
	if ($ranksA[$ii] > $ranksB[$ii]) { # a has better rank, i.e. comes first
	    #print "         : a\[$ii\] > b\[$ii\] => 1\n";
	    return 1;
	}
	elsif ($ranksA[$ii] < $ranksB[$ii]) {
	    #print "         : a\[$ii\] < b\[$ii\] => -1\n";
	    return -1;
	}
    }

    #print "         : no differences in ranks => 0\n";

    return 0; # no differences in the (comparable) ranks
}


sub cmpAllMaximum() {
    my @ranksA = reverse(@{&getAndSortRanks($a)});
    my @ranksB = reverse(@{&getAndSortRanks($b)}); 

    #print "comparing: a=$a; ranks: @ranksA\n";
    #print "         : b=$b; ranks: @ranksB\n";

    my $numRanks = scalar(@ranksA);
    if ($numRanks > scalar(@ranksB)) {
	$numRanks = scalar(@ranksB);
    }

    my $ii;
    for ($ii = 0; $ii < $numRanks; $ii++) {
	if ($ranksA[$ii] > $ranksB[$ii]) { # a has worse rank, i.e. comes first
	    #print "         : a\[$ii\] > b\[$ii\] => 1\n";
	    return 1;
	}
	elsif ($ranksA[$ii] < $ranksB[$ii]) {
	    #print "         : a\[$ii\] < b\[$ii\] => -1\n";
	    return -1;
	}
    }

    #print "         : no differences in ranks => 0\n";

    return 0; # no differences in the (comparable) ranks
}


sub cmpRank2ScoreProd() {
    if ($gene2score{$a} < $gene2score{$b}) {
	#print "comparison: $a=$gene2score{$a}; $b=$gene2score{$b} => 1\n";
	return 1;
    }

    if ($gene2score{$a} > $gene2score{$b}) {
	#print "comparison: $a=$gene2score{$a}; $b=$gene2score{$b} => -1\n";
	return -1;
    }

    # if both have the SAME probability, sort according to better ranking!
    #print "comparison: $a=$gene2score{$a}; $b=$gene2score{$b} => check ranks!\n";
    return &cmpAllMinimum($a,$b);
}

sub scoreAndWriteGenes() {

    my $numFiles = scalar(@RFILES);
    my %numScoredGenes = ();
    foreach my $RF (@RFILES) {
	$numScoredGenes{$RF} = scalar(keys(%{$gene2ranks{$RF}}));
    }

    %gene2score = ();

    my @genes = (); # for the final gene order according to scores, etc.

    if ($scoringFunct eq "allmaximum" || $scoringFunct eq "allminimum") {
	# handle this differently, because we need to directly compute
	# rankings, not some scores (average, rankprod, etc.)

	if ($scoringFunct eq "allmaximum") {
	    @genes = sort cmpAllMaximum @candidateGenes;
	}
	else {
	    @genes = sort cmpAllMinimum @candidateGenes;
	}

	my $rr;
	for ($rr=0; $rr < scalar(@genes); $rr++) {
	    $gene2score{$genes[$rr]} = $rr+1;
	}
    }
    else { # "average", "rankproduct", "rank2scoreproduct", "minimum", "maximum"

	foreach my $g (@candidateGenes) {

	    if($verboseOutput) {
		print "# computing score for gene $g:";
	    }

	    my $score = 0;
	    if ($scoringFunct eq "rankproduct") {
		$score = 1;
	    }
	    if ($scoringFunct eq "rank2scoreproduct") {
		$score = 1;
	    }
	    if ($scoringFunct eq "minimum") {
		$score = 999999999999;
	    }

	    my $numFilesWithScore = 0;

	    my $noScore = 0;
	    my $thisRank = 0;
	    foreach my $RF (@RFILES) {
		if (!exists($gene2ranks{$RF}{$g})) {
		    if ($strictScoring) {
			$noScore = 1;

			if($verboseOutput) {
			    print " warning for strict scoring: no rank for gene $g in file $RF!\n"; 
			}

			last;
		    }
		    next;
		}

		$numFilesWithScore++;

		if (!$rankProdScoring && $scoringFunct ne "rank2scoreproduct") {
		    # if not using rank product scores && not using score mapping: use relative rank instead of absolute rank!
		    $gene2ranks{$RF}{$g} /= $numScoredGenes{$RF};
		}
		$thisRank = $gene2ranks{$RF}{$g};

		if ($scoringFunct eq "average") {
		    $score += $gene2ranks{$RF}{$g};
		}
		elsif ($scoringFunct eq "rankproduct") {
		    $score *= $gene2ranks{$RF}{$g};
		}
		elsif ($scoringFunct eq "rank2scoreproduct") {
		              #[ choose the map for this result file... ][map the rank to score]
		    $score *= $r2sMaps{$rfid2rfname{$rfhandle2rfid{$RF}}}{$gene2ranks{$RF}{$g}};
		}
		elsif ($scoringFunct eq "maximum") {
		    if ($gene2ranks{$RF}{$g} > $score) {
			$score = $gene2ranks{$RF}{$g};
		    }
		}
		elsif ($scoringFunct eq "minimum") {
		    if ($gene2ranks{$RF}{$g} < $score) {
			$score = $gene2ranks{$RF}{$g};
		    }
		}
	    }

	    if ($noScore) {
		next;
	    }

	    if($verboseOutput) {
		print " rank: $thisRank, score now: $score;";
	    }

	    # final adjustments for the scores:
	    if ($scoringFunct eq "average") {
		# sum -> average
		$score /= $numFilesWithScore;
	    }
	    elsif ($scoringFunct eq "rankproduct" || $scoringFunct eq "rank2scoreproduct") {
		# normalize the rankproduct or score product according to the
		# number of files in which the gene was scored;
		# this leaves scores comparable if two genes have been
		# scored in different numbers of files.
		$score = $score**(1/$numFilesWithScore);
	    }
	    # nothing to do for maximum and minimum



	    $gene2score{$g} = $score;

	    if($verboseOutput) {
		print " => final $scoringFunct score over $numFilesWithScore files: $score\n";
	    }

	} # iteration over candidate genes $g ...

	# we got the scores, now sort in ascending order!
	@genes = keys(%gene2score);

	if ($scoringFunct eq "rank2scoreproduct") {
	    # from high to low prob. (upon draw from low to high rank)
	    @genes = sort cmpRank2ScoreProd @genes;
	}
	else {
	    # from lowest to higherst
	    @genes = sort {$gene2score{$a} <=> $gene2score{$b}} @genes;
	}

    } # "average", ""rankproduct", "rank2scoreproduct", "minimum", "maximum" ...



    my $currRank;
    my $writeRank; # need this to give two genes with the same score a shared,
                   # equal rank in the output file!
    my $currScore = -111;
    for ($currRank = 1; $currRank <= scalar(@genes); $currRank++) {

	if ($gene2score{$genes[$currRank-1]} != $currScore
	    || $scoringFunct eq "rank2scoreproduct") {
	    # the current gene has a different score, so adjust the rank to
	    # be written to this one; otherwise we write the same rank again
	    # (same score!)
	    # do also adjust the rank if we use rank2scoreproduct, where we
	    # can have equal probability(=score) but used ranks for further
	    # disambiguation!
	    $currScore = $gene2score{$genes[$currRank-1]};
	    $writeRank = $currRank;
	}

	# rank; gene; score_<merge>; 
	print "$writeRank\t$genes[$currRank-1]\t$gene2score{$genes[$currRank-1]}\t";
	# score_av;
	if ($scoringFunct eq "average") {
	    print "$gene2score{$genes[$currRank-1]}";
	}
	else {
	    print "-";
	}
	# relRank_scored; relRank_all
	print "\t".($writeRank/scalar(@genes))."\t".($writeRank/scalar(@candidateGenesAll));

	if ($extendedOutput) {
	    # single ranks
	    foreach my $RF (@RFILES) {
		if (exists($gene2ranks{$RF}{$genes[$currRank-1]})) {
		    print "\t$gene2ranks{$RF}{$genes[$currRank-1]}";
		}
		else {
		    print "\tunscored";
		}
	    }
	}
	print "\n";
    }

    # finally add all the unscored stuff!
    my $numGenesTotal = scalar(@candidateGenesAll);
    foreach my $g (@candidateGenesAll) {
	if (!exists($gene2score{$g})) {
	    # don't update the rank anymore ... not meaning full
	    #print "$currRank\t$g\tunscored\tunscored\t-\t".($currRank/scalar(@candidateGenesAll));
	    print "$numGenesTotal\t$g\tunscored\tunscored\t-\t-";

	    if ($extendedOutput) {
		foreach my $RF (@RFILES) {
		    if (exists($gene2ranks{$g})) {
			print "\t$gene2ranks{$g}";
		    }
		    else {
			print "\tunscored";
		    }
		}
	    }
	    print "\n";
	    # don't update the rank anymore ... not meaning full
	    #$currRank++;
	}
    }

    return;
}


sub writeRunAndInfoLines {

    # run information (OMIM ID and gene)
    foreach my $rL (@runLines) {
	print "$rL\n";
    }

    # average number of ref. genes for this OMIM
    print "\# $NUM_REF_4_PHENO $thisPhenotype: $numRefGenes4ThisPhenotype [average]\n";

    # average number of ref. genes
    print "\# $NUM_REF_EXPR: $numRefGenes [average]\n";

    # average of averages
    if ($avRefCorrScore ne "-") {
	$avRefCorrScore /= $numAvRefCorrScores;
    }
    print "\# $AV_PW_CORR: $avRefCorrScore [average]\n";

    # number of scored candidates; depends on what merging function we use!

    # number of candidate genes
    my $numCandUsed = 0;
    foreach my $g (@candidateGenes) {
 	my $noScore = 0;
	foreach my $RF (@RFILES) {
	    if (!exists($gene2ranks{$RF}{$g})) {
		if ($strictScoring) {
		    $noScore = 1;
		    last;
		}
		next;
	    }
	}
	if (!$noScore) {
	    $numCandUsed++;
	}
    }

    print "\# $NUM_CAN_SCORED$necData4Scoring: $numCandUsed\n";


    # other information lines plus "rank" header
    foreach my $iL (@infoLines) {
	print "$iL\n";
    }

    # after info lines, print a single header line
    print "$rankHeader";
    if ($extendedOutput) {
	foreach my $RF (@RFILES) {
	    if ($rankProdScoring) {
		print "\tscore_file$rfhandle2rfid{$RF}";
	    }
	    else {
		print "\trank_file$rfhandle2rfid{$RF}";
	    }
	}
    }
    print "\n";

    return;
}
