#!/usr/bin/perl -w

##############################################################################
# The purpose of this script is to run mbc-evalCandGenesByCoExpr over multiple
# genes that are known to be involved in disease (with OMIM phenotype IDs) to
# evaluate its performance (genes are first stripped of the annotation for the
# query phenotype to simulate a phenotype with unknown molecular basis).
# The considered gene and a specified number of adjacent genes (in both
# directions on the same chromosome) is taken as the candidate list to see
# what relative ranks the actual disease genes have in the prioritized
# candidate lists.
#
# copyright 2007 by Rosario M. Piro (rosario.piro@unito.it)
#
# ---------------------------------------------------------------------------
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; _exclusively_ version 2 of the License
#    NO later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
##############################################################################

use strict;

use Getopt::Std;
use File::Basename;

# command buffering of STDOUT
$| = 1;

my $prgname = "mbc-evalCandGenesByCoExpr-leaveOneOut.pl";


# for this script:
#my $omimLocusMapFile = "";
my $keepTmpFiles = 0;
my $removePhenoAnnoFromAllGenes = 0;

# for rmp-getENSGrange.pl
my $ensemblGeneListFile = "";
#my $maxNumCandidateGenes = 0;
my $numAdjacentGenes = 87; # default: dis. gene + 2x87 genes = 175 genes, which
                           # should correspond (more or less) to the median
                           # number of candidates in OMIM phenotypes with 
                           # associated genemap loci with but unknwon molecular
                           # basis (average of 109 according to Lage et al.
                           # Nature Biotechnology 25(3), 2007; but we found an
                           # average of about 260!)

my $phenotype2relatedGenesFile = "";
my $gene2OmimMapFile = "";

my $expressionFile = "";
my $scoringFunc = "";
my $coexprLists4CandGenes = 0;
my $dgpFile = "";

my $verboseOutput = 0;

my $tmpFileDir = "."; # directory for temporary files

my $candGenesTmpFile = "";
my $refGenesTmpFile = "";

my %ProcessedEnsgOmimIdPairs = ();

# -------------------
# getting options ...
# -------------------

my %opts;
$opts{h} = "";
$opts{k} = "";
$opts{t} = "";
$opts{g} = "";
$opts{p} = "";
$opts{m} = "";
$opts{e} = "";
$opts{s} = "";
$opts{C} = "";
$opts{d} = "";
$opts{n} = "";
$opts{R} = "";
$opts{v} = "";


getopts('hkt:g:p:m:e:s:Cd:n:Rv', \%opts);

if ( $opts{h} eq "1" ) {
    &printUsageInfo();
}

if ( $opts{t} ne "" ) {
    $tmpFileDir = $opts{t};
}

if ( $opts{p} ne "" ) {
    $phenotype2relatedGenesFile = $opts{p};
}

if ( $opts{k} eq "1" ) {
    $keepTmpFiles = 1;
}

if ( $opts{g} ne "" ) {
    $ensemblGeneListFile = $opts{g};
}

if ( $opts{m} ne "" ) {
    $gene2OmimMapFile = $opts{m};
}

if ( $opts{e} ne "" ) {
    $expressionFile = $opts{e};
}

if ( $opts{s} ne "" ) {
    $scoringFunc = $opts{s};
}

if ( $opts{C} eq "1" ) {
    $coexprLists4CandGenes = 1;
}

if ( $opts{d} ne "" ) {
    $dgpFile = $opts{d};
}

if ( $opts{n} =~ /^\d+$/ ) {
    $numAdjacentGenes = int($opts{n});
}

if ( $opts{R} ne "" ) {
    $removePhenoAnnoFromAllGenes = 1;
}

if ( $opts{v} eq "1" ) {
    $verboseOutput = 1;
}



if ( $ensemblGeneListFile eq "" || $phenotype2relatedGenesFile eq ""
     || $gene2OmimMapFile eq "" || $expressionFile eq "" ) {
    print STDERR "Error: At least one of the mandatory parameters is missing! (-h for help)\n";
    exit(1);
}

if (! -d $tmpFileDir) {
    print STDERR "Error: $tmpFileDir is not a directory!\n";
    exit(2);
}

if (! -e $ensemblGeneListFile || ! -r $ensemblGeneListFile) {
    print STDERR "Error: $ensemblGeneListFile does not exist or cannot be read!\n";
    exit(3);
}

if (! -e $phenotype2relatedGenesFile || ! -r $phenotype2relatedGenesFile) {
    print STDERR "Error: $phenotype2relatedGenesFile does not exist or cannot be read!\n";
    exit(3);
}

if (! -e $gene2OmimMapFile || ! -r $gene2OmimMapFile) {
    print STDERR "Error: $gene2OmimMapFile does not exist or cannot be read!\n";
    exit(3);
}

if (! -e $expressionFile || ! -r $expressionFile) {
    print STDERR "Error: $expressionFile does not exist or cannot be read!\n";
    exit(3);
}

if ($dgpFile ne "" && (! -e $dgpFile || ! -r $dgpFile)) {
    print STDERR "Error: $dgpFile does not exist or cannot be read!\n";
    exit(3);
}

#if (! -e "rmp-getENSGrange.pl" || ! -x "rmp-getENSGrange.pl") {
#    print STDERR "Error: rmp-getENSGrange.pl does not exist or cannot be executed!\n";
#    exit(4);
#}

#if (! -e "mbc-evalCandGenesByCoExpr" || ! -x "mbc-evalCandGenesByCoExpr") {
#    print STDERR "Error: mbc-evalCandGenesByCoExpr does not exist or cannot be executed!\n";
#    exit(4);
#}



# ----------------------------
# read and process input file:
# ----------------------------

unless ( open(OMIMIN, "< $gene2OmimMapFile") ) {
    print STDERR "Error: cannot open input file $gene2OmimMapFile!\n";
    exit(1);
}

if ($verboseOutput) {
    print "# Reading gene->OMIM phenotype map file ".basename($gene2OmimMapFile)." ...\n";
}

my $line = "";
my $lineNumber = 0;
while ($line = <OMIMIN>) {

    chomp($line);

    $lineNumber++;

    # comment or empty line?
    next if ($line =~ /^!!/ || $line =~ /^\#/ || $line =~ /^$/);

    my @lineTokens = split(/\t/, $line);
    if (scalar(@lineTokens) != 2) {
	print STDERR "Error: line $lineNumber of $gene2OmimMapFile contains a wrong number of\ncolumns. Got: ".scalar(@lineTokens)."; expected: 2. Quitting!\n";
	close(OMIMIN);
	exit(1);
    }

    my $geneID = $lineTokens[0];
    my $omimID = $lineTokens[1];

    print "+++ OMIM ID=$omimID gene=$geneID +++\n";

    my $timestamp = time();

    print "--- OMIM ID=$omimID gene=$geneID ---";

    # check if we already considered this OMIM ID/gene pair:
    if (exists($ProcessedEnsgOmimIdPairs{"$omimID $geneID"})) {
	print " (already processed ... skipping!)\n";
	next;
    }
    print "\n";


    # get reference genes for phenotype_query = $omimID
    # -------------------------------------------------

    $refGenesTmpFile = "$tmpFileDir/tmp_reference_genes_${omimID}_${geneID}_${timestamp}.dat";

    my $cmd = "grep \"^[^[:space:]][^[:space:]]*[[:space:]]${omimID}[[:space:]]\" $phenotype2relatedGenesFile";

    if ($removePhenoAnnoFromAllGenes) {
	# remove all reference genes directly involved in the
	# query disease (query phenotype = phenotype_gene)?
	#$cmd .= " | grep \"^[^[:space:]]*[[:space:]][^[:space:]]*[[:space:]]${omimID}[[:space:]]\" -v";

	$cmd .= " | uniqWithUnionOfValues.pl -k 1 | grep \"[^[:space:]]*[[:space:]][^[:space:]]*[[:space:]][^[:space:]]*${omimID}[^0-9]\" -v";
    }
    else {
	# otherwise, remove at least this disease gene from reference genes
	$cmd .= " | grep \"^${geneID}[[:space:]]\" -v";
    }

    


    $cmd .= " > $refGenesTmpFile";

    if ($verboseOutput) {
	print "# Executing: $cmd\n";
    }

    my $exitStatus = system($cmd);
    if ($exitStatus != 0) {
	print "Warning: No reference genes for this OMIM ID ($omimID)! Skipping ...\n";
    }
    else {
	# we have reference genes, continue!

	# now get candidate genes!
	# ------------------------

	$candGenesTmpFile = "$tmpFileDir/tmp_candidate_genes_${omimID}_${geneID}_${timestamp}.dat";
	#push(@tmpFiles, $candGenesTmpFile);

	$cmd = "bash -c 'rmp-getENSGrange.pl -i $ensemblGeneListFile -g $geneID -n $numAdjacentGenes > $candGenesTmpFile 2> /dev/null'";

	if ($verboseOutput) {
	    print "# Executing: $cmd\n";
	}

	$exitStatus = system($cmd);
	if ($exitStatus != 0) {
	    print "Error ($exitStatus) while executing: $cmd\n";
	    print "Quitting!\n";
	    exit(15);
	}

	if ( -z $candGenesTmpFile ) { # candidate gene file is empty!
	    print "Warning: No candidate genes for this artificial locus! Skipping ...\n";
	}
	else {
	    # we have also candidate genes, now evaluate the stuff!

	    # determine number of reference genes involved in this phenotype:
	    $cmd = "grep \"^[^[:space:]]*[[:space:]][0-9]*[[:space:]]${omimID}[[:space:]]\" $refGenesTmpFile";
	    if ($verboseOutput) {
		print "# Executing: $cmd\n";
	    }
	    my $grepResult = `$cmd`;
	    chomp ($grepResult);
	    my @refGenes4Pheno = split(/\n/, $grepResult);
	    my $numRefGenes4PhenoWithExprProfiles = 0;
	    foreach my $rG4P (@refGenes4Pheno) {
		if ($rG4P =~ /^([^\s]+)\s/) {
		    my $rGID = $1;
		    $cmd = "grep \"^${rGID}[[:space:]]\" $expressionFile > /dev/null";
		    if ($verboseOutput) {
			print "# Executing: $cmd\n";
		    }
		    if(!system($cmd)) {
			# has an expression profile
			$numRefGenes4PhenoWithExprProfiles++;
		    }
		}
	    }
	    print "# Number of reference genes with expression profile associated to OMIM ID $omimID: $numRefGenes4PhenoWithExprProfiles (of ".scalar(@refGenes4Pheno)." associated genes)\n";



	    # run mbc-evalCandGenesByCoExpr:

	    $cmd = "mbc-evalCandGenesByCoExpr -c $candGenesTmpFile -r $refGenesTmpFile -e $expressionFile";

	    if ($scoringFunc ne "") {
		$cmd .= " -s $scoringFunc";
	    }

	    if ($coexprLists4CandGenes) {
		$cmd .= " -C";
	    }

	    if ($dgpFile ne "") {
		$cmd .= " -d $dgpFile"
		}

	    if ($verboseOutput) {
		print "# Executing: $cmd\n";
	    }

	    $exitStatus = system($cmd);
	    if ($exitStatus != 0) {
		print "Error ($exitStatus) while executing: $cmd\n";
		print "Quitting!\n";
		exit(16);
	    }


	    $ProcessedEnsgOmimIdPairs{"$omimID $geneID"} = 1; # do not process twice

	} # else ... of if ( -z $candGenesTmpFile )   [having candidate genes]

    } # else ... of if ($exitStatus != 0)  [having reference genes]

    # remove temporary files?
    if (!$keepTmpFiles) {
	if ($verboseOutput) {
	    print "Removing temporary files '$candGenesTmpFile' and '$refGenesTmpFile' ...\n";
	}

	if (system("rm -f $candGenesTmpFile") != 0) {
	    print "Warning: cannot delete $candGenesTmpFile\n";
	}
	if (system("rm -f $refGenesTmpFile") != 0) {
	    print "Warning: cannot delete $refGenesTmpFile\n";
	}
    }

}


print "Summary: A total of ".scalar(keys(%ProcessedEnsgOmimIdPairs))." gene / OMIM phenotype ID pairs have been evaluated!\n";


close(OMIMIN);


#############################################################################
#                                   FUNCTIONS:
#############################################################################


# ---------------------------------------------------------------------------
#
# Print usage information (help).
# arguments: none
#
# --------------------------------------------------------------------------
sub printUsageInfo() {
    print "Author: Rosario M. Piro (rosario.piro\@unito.it)\n\n";
    print "Usage:\n";
    print "$prgname [OPTIONS]\n";
    print "\nWhere OPTIONS are:\n";
    print "-h                Display this help message and exit.\n";
    print "-k                Keep temporary files (may me many!).\n";
    print "-t <dir>          Directory for temporary files (default: working directory).\n";
    print "-g <genelistfile> The name of the file that contains the BioMart output\n";
    print "                  with the Ensembl gene list, that is already sorted\n";
    print "                  according to the gene locus (MANDATORY).\n";
    print "-n <num>          Number of adjacent genes of known disease gene to be\n";
    print "                  considered as candidate genes. Note: <num> genes from\n";
    print "                  both directions on the chromosome will be taken, hence the\n";
    print "                  total number of candidates will be 2x<num> + 1.\n";
    print "                  Default: $numAdjacentGenes.\n";
    print "-p <relgenefile>  The file containing the list of all genes associated to query\n";
    print "                  and related phenotypes. The file should have the same format\n";
    print "                  as the output of \"rmp-findDiseaseRelatedGenes.pl -O all\".\n";
    print "-m <e2omapfile>   The file containing a mapping from gene to OMIM IDs.\n";
    print "-R                Strip the OMIM phenotype annotation (regarding the query\n";
    print "                  disease) not only from the query gene (from <e2omapfile>),\n";
    print "                  but from all genes (in order to simulate a phenotype with\n";
    print "                  unknown molecular basis).\n";
    print "-e <exprfile>     File containing the expression profiles of all genes;\n";
    print "-s <func>         Scoring function to use for computing the co-expression of\n";
    print "                  two genes; see: \"mbc-evalCandGenesByCoExpr -h\"!\n";
    print "-C                Build co-expression lists for the candidate genes instead of\n";
    print "                  the reference genes (check ranks of reference genes within\n";
    print "                  the co-expression lists of candidate genes instead of\n";
    print "                  checking the ranks of candidate genes within the reference\n";
    print "                  genes' co-expression lists).\n";
    print "-v                Verbose stdout.\n\n";
    print "Mandatory arguments are: -g, -p, -m, -e.\n";

    exit(0);

} # printUsageInfo() ...

