#!/usr/bin/perl -w

##############################################################################
# The purpose of this script is to run mbc-evalCandGenesByCoExpr over multiple
# genes that are known to be involved in disease (with OMIM phenotype IDs) to
# evaluate its performance (genes are first stripped of the annotation for the
# query phenotype to simulate a phenotype with unknown molecular basis).
# The considered gene and a specified number of adjacent genes (in both
# directions on the same chromosome) is taken as the candidate list to see
# what relative ranks the actual disease genes have in the prioritized
# candidate lists.
#
# copyright 2007 by Rosario M. Piro (rosario.piro@unito.it)
#
# ---------------------------------------------------------------------------
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; _exclusively_ version 2 of the License
#    NO later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
##############################################################################

use strict;

use Getopt::Std;
use File::Basename;

# command buffering of STDOUT
$| = 1;

my $prgname = "mbc-evalCandGenesByCoExpr-multipleLoci.pl";


# for this script:
#my $omimLocusMapFile = "";
my $keepTmpFiles = 0;
#my $removePhenoAnnoFromAllGenes = 0;

# for rmp-getENSGrange.pl
my $ensemblGeneListFile = "";

my $phenotype2relatedGenesFile = "";
#my $gene2OmimMapFile = "";

my $omimLociFile = "";

my $expressionFile = "";
my $scoringFunc = "";
my $coexprLists4CandGenes = 0;
my $dgpFile = "";

my $extendedOutputFlag = "";  # default: no extended output
my $verboseOutput = 0;

my $tmpFileDir = "."; # directory for temporary files

my $candGenesTmpFile = "";
my $refGenesTmpFile = "";

my %ProcessedOmimIdLocusPairs = ();

# -------------------
# getting options ...
# -------------------

my %opts;
$opts{h} = "";
$opts{k} = "";
$opts{t} = "";
$opts{g} = "";
$opts{p} = "";
$opts{e} = "";
$opts{s} = "";
$opts{C} = "";
$opts{d} = "";
$opts{E} = "";
$opts{o} = "";
$opts{v} = "";


getopts('hkt:g:p:e:s:Cd:o:Ev', \%opts);

if ( $opts{h} eq "1" ) {
    &printUsageInfo();
}

if ( $opts{t} ne "" ) {
    $tmpFileDir = $opts{t};
}

if ( $opts{p} ne "" ) {
    $phenotype2relatedGenesFile = $opts{p};
}

if ( $opts{k} eq "1" ) {
    $keepTmpFiles = 1;
}

if ( $opts{g} ne "" ) {
    $ensemblGeneListFile = $opts{g};
}

#if ( $opts{m} ne "" ) {
#    $gene2OmimMapFile = $opts{m};
#}

if ( $opts{e} ne "" ) {
    $expressionFile = $opts{e};
}

if ( $opts{s} ne "" ) {
    $scoringFunc = $opts{s};
}

if ( $opts{C} eq "1" ) {
    $coexprLists4CandGenes = 1;
}

if ( $opts{d} ne "" ) {
    $dgpFile = $opts{d};
}

if ( $opts{o} ne "" ) {
    $omimLociFile = $opts{o};
}

#if ( $opts{R} ne "" ) {
#    $removePhenoAnnoFromAllGenes = 1;
#}

if ( $opts{E} eq "1" ) {
    $extendedOutputFlag = "-E";
}

if ( $opts{v} eq "1" ) {
    $verboseOutput = 1;
}



if ( $ensemblGeneListFile eq "" || $phenotype2relatedGenesFile eq ""
     || $omimLociFile eq "" || $expressionFile eq "" ) {
    print STDERR "Error: At least one of the mandatory parameters is missing! (-h for help)\n";
    exit(1);
}

if (! -d $tmpFileDir) {
    print STDERR "Error: $tmpFileDir is not a directory!\n";
    exit(2);
}

if (! -e $ensemblGeneListFile || ! -r $ensemblGeneListFile) {
    print STDERR "Error: $ensemblGeneListFile does not exist or cannot be read!\n";
    exit(3);
}

if (! -e $phenotype2relatedGenesFile || ! -r $phenotype2relatedGenesFile) {
    print STDERR "Error: $phenotype2relatedGenesFile does not exist or cannot be read!\n";
    exit(3);
}

if (! -e $omimLociFile || ! -r $omimLociFile) {
    print STDERR "Error: $omimLociFile does not exist or cannot be read!\n";
    exit(3);
}

if (! -e $expressionFile || ! -r $expressionFile) {
    print STDERR "Error: $expressionFile does not exist or cannot be read!\n";
    exit(3);
}

if ($dgpFile ne "" && (! -e $dgpFile || ! -r $dgpFile)) {
    print STDERR "Error: $dgpFile does not exist or cannot be read!\n";
    exit(3);
}

#if (! -e "rmp-getENSGrange.pl" || ! -x "rmp-getENSGrange.pl") {
#    print STDERR "Error: rmp-getENSGrange.pl does not exist or cannot be executed!\n";
#    exit(4);
#}

#if (! -e "mbc-evalCandGenesByCoExpr" || ! -x "mbc-evalCandGenesByCoExpr") {
#    print STDERR "Error: mbc-evalCandGenesByCoExpr does not exist or cannot be executed!\n";
#    exit(4);
#}



# ----------------------------
# read and process input file:
# ----------------------------

unless ( open(OMIMIN, "< $omimLociFile") ) {
    print STDERR "Error: cannot open input file $omimLociFile!\n";
    exit(1);
}

if ($verboseOutput) {
    print "# Reading OMIM phenotype -> locus map file ".basename($omimLociFile)." ...\n";
}

my $line = "";
my $lineNumber = 0;
while ($line = <OMIMIN>) {

    chomp($line);

    $lineNumber++;

    # comment or empty line?
    next if ($line =~ /^!!/ || $line =~ /^\#/ || $line =~ /^$/);

    my @lineTokens = split(/\t/, $line);
    if (scalar(@lineTokens) != 3) {
	print STDERR "Error: line $lineNumber of $omimLociFile contains a wrong number of\ncolumns. Got: ".scalar(@lineTokens)."; expected: 3. Quitting!\n";
	close(OMIMIN);
	exit(1);
    }

    my $omimID = $lineTokens[0];
    my $omimLocusList = $lineTokens[1];
    my $omimDescr = $lineTokens[2];

    # remove initial class symbol (or any non-digit character) from OMIM ID:
    $omimID =~ s/^\D//g;

    if ($omimID !~ /^\d+$/) {
	print STDERR "Error: cannot parse OMIM phenotype ID in line $lineNumber of ".basename($omimLociFile).": '$omimID'\n";
	exit(11);
    }

    print "+++ OMIM ID=$omimID ($omimDescr) locus=$omimLocusList +++\n";

    my $timestamp = time();

    # check if the locus is present:
    if ($omimLocusList eq "(none)" || $omimLocusList eq "") {
	print " (locus not defined ... skipping!)\n";
	next;
    }

    $omimLocusList =~ s/ //g;  # remove all spaces if there are any.
    my @locusVect = split(/,/, $omimLocusList);

    foreach my $omimLocus (@locusVect) {

	print "--- OMIM ID=$omimID ($omimDescr) locus=$omimLocus ---\n";

	# check if we already considered this OMIM ID/gene pair:
	if (exists($ProcessedOmimIdLocusPairs{"$omimID $omimLocus"})) {
	    print " (already processed ... skipping!)\n";
	    next;
	}


	# get reference genes for phenotype_query = $omimID
	# -------------------------------------------------

	$refGenesTmpFile = "$tmpFileDir/tmp_reference_genes_${omimID}_${timestamp}.dat";

	my $cmd = "grep \"^[^[:space:]][^[:space:]]*[[:space:]]${omimID}[[:space:]]\" $phenotype2relatedGenesFile";

	# this was for the leave-one-out and should serve no purpose here:
	#if ($removePhenoAnnoFromAllGenes) {
	  ## remove all reference genes directly involved in the
	  ## query disease (phenotype_query = phenotype_gene)?
	  #$cmd .= " | grep \"[[:space:]]${omimID}[[:space:]]${omimID}[[:space:]]\" -v";
    #}
	#else {
	  ## otherwise, remove at least this disease gene from reference genes
	  #$cmd .= " | grep \"^${geneID}[[:space:]]\" -v";
	#}

	$cmd .= " > $refGenesTmpFile";

	if ($verboseOutput) {
	    print "# Executing: $cmd\n";
	}

	my $exitStatus = system($cmd);
	if ($exitStatus != 0) {
	    print "Warning: No reference genes for this OMIM ID ($omimID)! Skipping ...\n";
	}
	else {
	    # we have reference genes, continue!

	    # now get candidate genes!
	    # ------------------------

	    $candGenesTmpFile = "$tmpFileDir/tmp_candidate_genes_${omimID}_${timestamp}.dat";
	    #push(@tmpFiles, $candGenesTmpFile);

	    $cmd = "rmp-getENSGrange.pl -i $ensemblGeneListFile -l $omimLocus > $candGenesTmpFile";

	    if ($verboseOutput) {
		print "# Executing: $cmd\n";
	    }

	    $exitStatus = system($cmd);
	    if ($exitStatus != 0) {
		print "Error ($exitStatus) while executing: $cmd\n";
		print "Quitting!\n";
		exit(15);
	    }

	    if ( -z $candGenesTmpFile ) { # candidate gene file is empty!
		print "Warning: No candidate genes for this artificial locus! Skipping ...\n";
	    }
	    else {
		# we have also candidate genes, now evaluate the stuff!

		# determine number of reference genes involved in this phenotype:
		$cmd = "grep \"^[^[:space:]]*[[:space:]][0-9]*[[:space:]]${omimID}[[:space:]]\" $refGenesTmpFile | extractSpecColumns.pl -c 1 | sort | uniq";
		if ($verboseOutput) {
		    print "# Executing: $cmd\n";
		}
		my $grepResult = `$cmd`;
		chomp ($grepResult);
		my @refGenes4Pheno = split(/\n/, $grepResult);
		my $numRefGenes4PhenoWithExprProfiles = 0;
		foreach my $rG4P (@refGenes4Pheno) {
		    $cmd = "grep \"^${rG4P}[[:space:]]\" $expressionFile > /dev/null";
		    if ($verboseOutput) {
			print "# Executing: $cmd\n";
		    }
		    if(!system($cmd)) {
			# has an expression profile
			$numRefGenes4PhenoWithExprProfiles++;
		    }
		}
		print "# Number of reference genes with expression profile associated to OMIM ID $omimID: $numRefGenes4PhenoWithExprProfiles (of ".scalar(@refGenes4Pheno)." associated genes)\n";


		# run mbc-evalCandGenesByCoExpr:

		$cmd = "mbc-evalCandGenesByCoExpr -c $candGenesTmpFile -r $refGenesTmpFile -e $expressionFile $extendedOutputFlag";

		if ($scoringFunc ne "") {
		    $cmd .= " -s $scoringFunc";
		}

		if ($coexprLists4CandGenes) {
		    $cmd .= " -C";
		}

		if ($dgpFile ne "") {
		    $cmd .= " -d $dgpFile"
		    }

		if ($verboseOutput) {
		    print "# Executing: $cmd\n";
		}

		$exitStatus = system($cmd);
		if ($exitStatus != 0) {
		    print "Error ($exitStatus) while executing: $cmd\n";
		    print "Quitting!\n";
		    exit(16);
		}


		$ProcessedOmimIdLocusPairs{"$omimID $omimLocus"} = 1; # do not process twice

	    } # else ... of if ( -z $candGenesTmpFile )   [having candidate genes]

	} # else ... of if ($exitStatus != 0)  [having reference genes]

	# remove temporary files?
	if (!$keepTmpFiles) {
	    if ($verboseOutput) {
		print "Removing temporary files '$candGenesTmpFile' and '$refGenesTmpFile' ...\n";
	    }

	    if (system("rm -f $candGenesTmpFile") != 0) {
		print "Warning: cannot delete $candGenesTmpFile\n";
	    }
	    if (system("rm -f $refGenesTmpFile") != 0) {
		print "Warning: cannot delete $refGenesTmpFile\n";
	    }
	}

    } # foreach my $omimLocus ...

}


print "Summary: A total of ".scalar(keys(%ProcessedOmimIdLocusPairs))." OMIM phenotype / locus pairs have been prioritized!\n";


close(OMIMIN);


#############################################################################
#                                   FUNCTIONS:
#############################################################################


# ---------------------------------------------------------------------------
#
# Print usage information (help).
# arguments: none
#
# --------------------------------------------------------------------------
sub printUsageInfo() {
    print "Author: Rosario M. Piro (rosario.piro\@unito.it)\n\n";
    print "Usage:\n";
    print "$prgname [OPTIONS]\n";
    print "\nWhere OPTIONS are:\n";
    print "-h                Display this help message and exit.\n";
    print "-k                Keep temporary files (may me many!).\n";
    print "-t <dir>          Directory for temporary files (default: working directory).\n";
    print "-o <omimlocifile> The name of the file that contains the list of orphan loci\n";
    print "                  from OMIM for which to evaluate positional candidates.\n";
    print "                  Required format: \"<omimPhenotypeID>\\t<locus>\\t<descr>\"\n";
    print "-g <genelistfile> The name of the file that contains the BioMart output\n";
    print "                  with the Ensembl gene list, that is already sorted\n";
    print "                  according to the gene locus (MANDATORY).\n";
    print "-p <relgenefile>  The file containing the list of all genes associated to query\n";
    print "                  and related phenotypes. The file should have the same format\n";
    print "                  as the output of \"rmp-findDiseaseRelatedGenes.pl -O all\".\n";
#    print "-m <e2omapfile>   The file containing a mapping from gene to OMIM IDs.\n";
#    print "-R                Strip the OMIM phenotype annotation (regarding the query\n";
#    print "                  disease) not only from the query gene (from <e2omapfile>),\n";
#    print "                  but from all genes (in order to simulate a phenotype with\n";
#    print "                  unknown molecular basis).\n";
    print "-e <exprfile>     File containing the expression profiles of all genes;\n";
    print "-s <func>         Scoring function to use for computing the co-expression of\n";
    print "                  two genes; see: \"mbc-evalCandGenesByCoExpr -h\"!\n";
    print "-C                Build co-expression lists for the candidate genes instead of\n";
    print "                  the reference genes (check ranks of reference genes within\n";
    print "                  the co-expression lists of candidate genes instead of\n";
    print "                  checking the ranks of candidate genes within the reference\n";
    print "                  genes' co-expression lists).\n";
    print "-E                Extended output (relative positions in co-expression lists).\n";
    print "-v                Verbose stdout.\n\n";
    print "Mandatory arguments are: -o, -g, -p, -e.\n";

    exit(0);

} # printUsageInfo() ...

