#!/usr/bin/perl -w

##############################################################################
# The purpose of this script is to summarize the results contained in the
# output file of mbc-evalCandGenesByCoExpr-leaveOneOut.pl
#
# copyright 2007-2010 by Rosario M. Piro (rosario.piro@unito.it)
#
# ---------------------------------------------------------------------------
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; _exclusively_ version 2 of the License
#    NO later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
##############################################################################

use strict;

use Getopt::Std;
use File::Basename;
use POSIX qw(strtod);

# command buffering of STDOUT
$| = 1;

my $prgname = "mbc-evalCandGenesByCoExpr-reorderByAvScore.pl";

my $resultsFile = "";
my $verboseOutput = 0;



# -------------------
# getting options ...
# -------------------

my %opts;
$opts{h} = "";
$opts{r} = "";
$opts{v} = "";

getopts('hr:v', \%opts);

if ( $opts{h} eq "1" ) {
    &printUsageInfo();
}

if ( $opts{r} ne "" ) {
    $resultsFile = $opts{r};
}

if ( $opts{v} eq "1" ) {
    $verboseOutput = 1;
}

# ---------------
# checking files:
# ---------------

if ( $resultsFile eq "") {
    print STDERR "Error: At least one of the mandatory parameters is missing! (-h for help)\n";
    exit(1);
}

if (! -e $resultsFile || ! -r $resultsFile) {
    print STDERR "Error: $resultsFile does not exist or cannot be read!\n";
    exit(2);
}


# ----------------------------
# read and process input file:
# ----------------------------

unless ( open(RESIN, "< $resultsFile") ) {
    print STDERR "Error: cannot open input file $resultsFile!\n";
    exit(1);
}

if ($verboseOutput) {
    print "# Reading input file ".basename($resultsFile)." ...\n";
}

my $line = "";
my $lineNum = 0;

my @geneorder_scored = ();
my @geneorder_unscored = ();
my %gene2score_mult = ();
my %gene2score_av = ();

while ($line = <RESIN>) {

    $lineNum++;

    chomp($line);

    if ($line =~ /^\+\+\+/) {
	# new section of a result file, print the reordering for the previous
	# section before continuing ...

	&reorderByAvScore();
	&printResultsWithNewOrder();

	# start a new ranking
	@geneorder_scored = ();
	@geneorder_unscored = ();
	%gene2score_mult = ();
	%gene2score_av = ();
    }

    if ($line =~ /^#/ || $line =~ /^\+\+\+/ || $line =~ /^---/) {
	# comment, just print it

	print "$line\n";
    }
    elsif ($line =~ /^rank/) {
	# header, print only the first 6 column names

	if ($verboseOutput) {
	    print "# original header: $line\n";
	}

	my @cols = split (/\t/, $line);
	if (scalar(@cols) == 6) {
	    print "$line\n";
	}
	elsif (scalar(@cols) > 6) {
	    my $ii;
	    for ($ii = 0; $ii < 6; $ii++) {
		if ($ii != 0) {
		    print "\t";
		}
		print "$cols[$ii]";
	    }
	    print "\n";
	}
	else {
	    print STDERR "Error in line $lineNum: require at least 6 fields, having: ".scalar(@cols)."!\n";
	    exit(111);
	}
    }
    else {
	# this is one of the ranked genes, get info for re-ranking:

	if ($verboseOutput) {
	    print "# original rank: $line\n";
	}

	my @fields = split(/\t/, $line);
	if (scalar(@fields) < 6) {
	    print STDERR "Error in line $lineNum: require at least 6 fields, having: ".scalar(@fields)."!\n";
	    exit(111);
	}

	# field 0: old rank, not needed
	my $gene = $fields[1];
	$gene2score_mult{$gene} = $fields[2];
	$gene2score_av{$gene} = $fields[3];
	# field 4: old relRank_scored, not needed
	# field 5: old relRank_all, not needed

	if ($gene2score_av{$gene} =~ /^-?[\.0-9]+$/) {
	    # some numeric value, must be a score
	    push(@geneorder_scored, $gene);
	}
	else {
	    push(@geneorder_unscored, $gene);
	}
    }
}

# reorder and write also the final section (there is no "+++" after it)

&reorderByAvScore();
&printResultsWithNewOrder();


close(RESIN);

exit(0);

#############################################################################
#                                   FUNCTIONS:
#############################################################################


# ---------------------------------------------------------------------------
#
# Print usage information (help).
# arguments: none
#
# --------------------------------------------------------------------------
sub printUsageInfo() {
    print "Author: Rosario M. Piro (piro\@to.infn.it)\n\n";
    print "Usage:\n";
    print "$prgname [OPTIONS]\n";
    print "\nWhere OPTIONS are:\n";
    print "-h                Display this help message and exit.\n";
    print "-r <infile>       File containing the prioritization results (output of\n";
    print "                  mbc-evalCandGenesByCoExpr-leaveOneOut.pl).\n";
    print "-v                Verbose stdout.\n";
    print "Mandatory arguments are: -r.\n\n";

    exit(0);

} # printUsageInfo() ...


sub reorderByAvScore() {

    @geneorder_scored =
	sort {$gene2score_av{$a} <=> $gene2score_av{$b}} @geneorder_scored;

} # reorderByAvScore() ...


sub printResultsWithNewOrder() {

    my $rank = 0;

    # first the scored genes:
    foreach my $gene(@geneorder_scored) {
	$rank++;
	my $relRank_scored = $rank / scalar(@geneorder_scored);
	my $relRank_all = $rank /
	    (scalar(@geneorder_scored) + scalar(@geneorder_unscored));
	print "$rank"."\t";
	print "$gene"."\t";
	print "$gene2score_mult{$gene}"."\t";
	print "$gene2score_av{$gene}"."\t";
	print "$relRank_scored"."\t";
	print "$relRank_all"."\n";
    }

    # now print all unscored genes in the original order
    foreach my $gene(@geneorder_unscored) {
	$rank++;
	my $relRank_scored = "-";
	my $relRank_all = $rank /
	    (scalar(@geneorder_scored) + scalar(@geneorder_unscored));
	print "$rank"."\t";
	print "$gene"."\t";
	print "$gene2score_mult{$gene}"."\t";
	print "$gene2score_av{$gene}"."\t";
	print "$relRank_scored"."\t";
	print "$relRank_all"."\n";
    }

} # printResultsWithNewOrder() ...
