/*****************************************************************************
 * mbc-evalCandGenesByCoExpr-leaveOneOut.cpp - CGPACE - evaluation of        *
 * candidate genes, given a list of candidates, a list of known disease and  *
 * a list of Pearson correlation coefficients (or a similar measure for      *
 * co-expression; having higher values for positive co-expression).          *
 * Version 0.0.6                                                             *
 * ------------------------------------------------------------------------- *
 * Copyright 2007-2012 by Rosario Piro (mailto: rmpiro@gmail.com)            *
 *                                                                           *
 *   This program is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation; _exclusively_ version 2 of the License    *
 *   NO later version.                                                       *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program; if not, write to the Free Software             *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA *
 *   or see http://www.gnu.org/copyleft/gpl.html                             *
 * ------------------------------------------------------------------------- */

//---------------------
//--- include files ---
//---------------------
//#include <fstream>
#include <iostream>
//#include <map>
//#include <set>
//#include <vector>
#include <iomanip>
//#include <cmath>
#include <cstring>
#include <algorithm>
//#include <stdlib.h>
//#include <values.h>
#include "extstdlib.h"
#include "evalCandGenesByCoExpr-serviceroutines.h"
//#include <time.h>

//---------------------------
//--- preprocessor macros ---
//---------------------------
#define PRGVERSION "0.0.6"
#define DEF_NUM_ADJ_GENES 100

#define NO_EXPRVALUE DBL_MIN


//-----------------------------/
//--- function declarations ---/
//-----------------------------/

void printUsageInfo(const char *prgname);
int getCandGenesFromChrGeneList(vector<pair<string, string> > &chrGeneList,
				map<string,unsigned int> &gene2ListPos,
				const string thisDisGene,
				const unsigned int numAdjacentGenes,
				vector<string> &candGenes);

//------------------------/
//--- global variables ---/
//------------------------/

//------------
//--- main ---
//------------

int main (int argc, const char **argv) {

  //cerr << "TODO: implement usage of DGP scores!" << endl << flush;

  int exitStatus = 0;


  unsigned int numAdjacentGenes = DEF_NUM_ADJ_GENES;
  bool removePhenoAnnoFromAllGenes = false;
  string ensemblGeneListFile = "";
  string phenotype2relatedGenesFile = "";
  string gene2OmimMapFile = "";
  set<string> disGenes4L1O;


  bool useDGPprob = false;     // default
  bool extendedOutput = false; // default
  bool coexprLists4CandGenes = false; // default

  vector<string> headerInfo;
  map<string,vector<unsigned int> > planes;
  void *functSpecInfo = NULL;
  double minNormExpr = DBL_MIN; // used only for "relintoverlap"; default: no
                            // minimum/threshold.

  vector<pair<string, string> > chrGeneList; // ordered list of chr/gene-pairs
  map<string,unsigned int> gene2ListPos;     // gene positions in the list


  map<string, set<string> > omim2SimPhenoMap; // OMIM -> similar OMIMs
  map<string, set<string> > omim2GenesMap;    // OMIM -> disease genes

  set<string> genesWithExprData;

  map<string,vector<double> > exprProfileMap;

  vector<double> weights; // default: empty, no weighting applied

  double (*computeCoExprTwoGenes)(vector<double> &,
				  vector<double> &,
				  vector<double> &,
				  void *) = computePearson;
                               // points to the function that determins
                               // the co-expression of two genes.
                               // Default: Pearson correlation coefficient
  string coExprScoringFunction = "pearson";

  map<string,double> geneDGPmap; // maps genes to their DGP value
                                 // (Lopez-Bigas et al.)

  string exprDataFileName = "";
  string weightsFileName = "";
  string dgpFileName = "";


  char *dateString;

  cout << setprecision(DEFAULT_PRECISION);

  // parse command line arguments:
  int index = 1;
  while (index < argc) {
    // help
    if (strcmp(argv[index],"--help")==0 || strcmp(argv[index],"-h")==0) {
      printUsageInfo(argv[0]);
      exit(0);
    }
    // ensemblGeneListFile
    else if (strcmp(argv[index],"-g")==0) {
      if (index + 1 < argc) {
	ensemblGeneListFile = argv[++index];
      }
      else { // filename missing after -g
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    // Weight file
    else if (strcmp(argv[index],"-w")==0) {
      if (index + 1 < argc) {
	weightsFileName = argv[++index];
      }
      else { // filename missing after -w
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    // gene2OmimMapFile
    else if (strcmp(argv[index],"-m")==0) {
      if (index + 1 < argc) {
	gene2OmimMapFile = argv[++index];
      }
      else { // filename missing after -m
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    // phenotype2relatedGenesFile
    else if (strcmp(argv[index],"-p")==0) {
      if (index + 1 < argc) {
	phenotype2relatedGenesFile = argv[++index];
      }
      else { // filename missing after -p
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    /*
    // candidate gene file
    else if (strcmp(argv[index],"-c")==0) {
      if (index + 1 < argc) {
	candGeneFileName = argv[++index];
      }
      else { // filename missing after -c
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    // reference gene file
    else if (strcmp(argv[index],"-r")==0) {
      if (index + 1 < argc) {
	refGeneFileName = argv[++index];
      }
      else { // filename missing after -r
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    */
    // number of adjacent genes for artifical loci
    else if (strcmp(argv[index],"-n")==0) {
      if (index + 1 < argc) {
	numAdjacentGenes = atoi(argv[++index]);
      }
      else { // filename missing after -n
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    // Expression data file
    else if (strcmp(argv[index],"-e")==0) {
      if (index + 1 < argc) {
	exprDataFileName = argv[++index];
      }
      else { // filename missing after -e
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    // coexpression score for two genes:
    else if (strcmp(argv[index],"-s")==0) {
      if (index + 1 < argc) {
	++index;
	if (strcmp(argv[index],"pearson")==0) {
	  computeCoExprTwoGenes = computePearson;
	  coExprScoringFunction = "pearson";
	}
	else if (strcmp(argv[index],"quadraticpearson")==0) {
	  computeCoExprTwoGenes = computeQuadraticPearson;
	  coExprScoringFunction = "quadraticpearson";
	}
	else if (strcmp(argv[index],"euclidean")==0) {
	  computeCoExprTwoGenes = computeInverseEucDist;
	  coExprScoringFunction = "euclidean";
	}
	else if (strcmp(argv[index],"cosine")==0) {
	  computeCoExprTwoGenes = computeCosine;
	  coExprScoringFunction = "cosine";
	}
	else if (strcmp(argv[index],"spearman")==0) {
	  computeCoExprTwoGenes = computePearson;  // ! spearman = pearson on
	                                           //   rank transformed data!
	  coExprScoringFunction = "spearman";
	}
	else if (strcmp(argv[index],"relintoverlap")==0) {
	  computeCoExprTwoGenes = computeRelativeIntensityOverlap;
	  coExprScoringFunction = "relintoverlap";
	}
	//else if (strcmp(argv[index],"normpearson")) {
	//computeCoExprTwoGenes = computeNormalizedPearson;
	//coExprScoringFunction = "normpearson";
	//}
	else if (strcmp(argv[index],"pearson3d")==0) {
	  computeCoExprTwoGenes = computePearson3D;
	  coExprScoringFunction = "pearson3d";
	}
	else { // unknwon scoring function
	  printUsageInfo(argv[0]);
	  exit(1);
	}
      }
      else { // scoring function missing after -s
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    // build coexpression lists for candidates instead of reference genes? 
    else if (strcmp(argv[index],"-C")==0) {
      coexprLists4CandGenes = true;
    }
    // use minimum normalized expression for "relintoverlap"
    else if (strcmp(argv[index],"-M")==0) {
      if (index + 1 < argc) {
	const char *val = argv[++index];
	if (isNumeric(val))
	  minNormExpr = atof(val);
	else {
	  printUsageInfo(argv[0]);
	  exit(1);
	}
      }
      else { // threshold missing after -M
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    // output file
    //else if (strcmp(argv[index],"-o")==0) {
    //  if (index + 1 < argc) {
    //outputFileName = argv[++index];
    //}
    //else { // filename missing after -o
    //printUsageInfo(argv[0]);
    //exit(1);
    //}
    //}
    // use DGP probabilites
    else if (strcmp(argv[index],"-d")==0) {
      if (index + 1 < argc) {
	dgpFileName = argv[++index];
	useDGPprob = true;
      }
      else { // filename missing after -d
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    // extended output
    else if (strcmp(argv[index],"-E")==0) {
      extendedOutput = true;
    }
    // pretend phenotype with unknown molecular basis (no associated
    // disease genes)
    else if (strcmp(argv[index],"-R")==0) {
      removePhenoAnnoFromAllGenes = true;
    }

    index++;
  }


  // check for mandatory parameters:
  if (ensemblGeneListFile == "" || gene2OmimMapFile == ""
      || exprDataFileName == "" || phenotype2relatedGenesFile == "") {
    printUsageInfo(argv[0]);
    exit(1);
  }


  // NOT NEEDED ANYMORE // check the presence of external tools
  //string dummyHelp;
  //if (getStdOutFromCommand("rmp-getENSGrange.pl -h",
  //			   dummyHelp) != 0) {
  //  cerr << "Error: external tool rmp-getENSGrange.pl required, but not in path! Not installed?" << endl;
  //  exit(79);
  //}
  //if (getStdOutFromCommand("extractSpecColumns.pl -h", dummyHelp) != 0) {
  //  cerr << "Error: external tool extractSpecColumns.pl required, but not in path! Not installed?" << endl;
  //  exit(79);
  //}



  // before starting determine date and time:
  time_t currTime = time(NULL);
  dateString = ctime(&currTime);


  if (weightsFileName != "") {
    // read input file for weight vector:
    #ifdef DEBUG
    cout << "#DEBUG: [========== File: weight vector ==========] " << endl;
    #endif

    if ( (exitStatus=parseWeightFile(weightsFileName, weights)) != 0) {
      cerr << "Error: Couldn't read file " << weightsFileName
	   << " ... quitting! Errorcode: " << exitStatus << endl;
      exit(exitStatus);
    }
  }


  // read file mapping genes to OMIM IDs (similar and true)
  #ifdef DEBUG
  cout << "#DEBUG: [========== File: mapping gene to OMIM IDs (similar and true) ==========] " << endl;
  #endif
  if ( (exitStatus=parseGene2OmimMapFile(phenotype2relatedGenesFile,
					 omim2SimPhenoMap,
					 omim2GenesMap) != 0)) {
    cerr << "Error: Couldn't read file " << phenotype2relatedGenesFile
	 << " ... quitting! Errorcode: " << exitStatus << endl;
    exit(exitStatus);
  }

  // read file mapping genes to OMIM IDs (list for leave-one-out)
  #ifdef DEBUG
  cout << "#DEBUG: [========== File: mapping gene to OMIM IDs (list for leave-one-out) ==========] " << endl;
  #endif
  if ( (exitStatus=parseDisGeneListFile(gene2OmimMapFile,
					disGenes4L1O) != 0)) {
    cerr << "Error: Couldn't read file " << gene2OmimMapFile
	 << " ... quitting! Errorcode: " << exitStatus << endl;
    exit(exitStatus);
  }


  // read the ordered chromosome - gene list
  #ifdef DEBUG
  cout << "#DEBUG: [========== File: ordered Ensembl gene list ==========] " << endl;
  #endif
  if ( (exitStatus=parseGeneListFile(ensemblGeneListFile, chrGeneList,
				     gene2ListPos))) {
    cerr << "Error: Couldn't read file " << ensemblGeneListFile
	 << " ... quitting! Errorcode: " << exitStatus << endl;
    exit(exitStatus);
  }


  // read expression data file (expression profiles):
  #ifdef DEBUG
  cout << "#DEBUG: [========== File: Expression data ==========] " << endl;
  #endif

  if ( (exitStatus=parseExpressionDataFile(exprDataFileName, genesWithExprData,
					   exprProfileMap, headerInfo)) != 0) {
    cerr << "Error: Couldn't read file " << exprDataFileName
	 << " ... quitting! Errorcode: " << exitStatus << endl;
    exit(exitStatus);
  }

  // set additional input information or perform additional pre-processing
  // for some of the correlation measures:

  // if needed, parse header info to extract planes:
  if (coExprScoringFunction == "pearson3d") {
    if ((exitStatus=convertHeaderInfo2Planes(headerInfo, planes)) != 0) {
      cerr << "Error: Couldn't convert header info to planes "
	   << " ... quitting! Errorcode: " << exitStatus << endl;
      exit(exitStatus);
    }

    functSpecInfo = &planes; // need this to pass
  }
  else if (coExprScoringFunction == "relintoverlap") {
    // if we don't need planes, we may need a minimum normalized expr.
    // (for noise reduction with relintoverlap)
    functSpecInfo = &minNormExpr;
  }
  else if (coExprScoringFunction == "spearman") {
    // Spearman rank correlation = Pearson correlation on rank transformed data
    #ifdef DEBUG
    cout << "#DEBUG: [========== Rank transformation of expression data for Spearman ==========] " << endl;
    #endif

    
    map<string,vector<double> >::iterator ePMit = exprProfileMap.begin();
    unsigned int ePlength = ePMit->second.size();

    vector<pair<unsigned int, double> > idx2valueVect;
    idx2valueVect.resize(ePlength);


    for ( ; ePMit != exprProfileMap.end(); ePMit++) {
      // for each expression profile ...

      #ifdef DEBUG
      cout << "#DEBUG: expression profile for " << ePMit->first << endl;
      cout << "#DEBUG:    profile:";
      for (vector<double>::iterator vit = ePMit->second.begin();
	   vit != ePMit->second.end(); vit++)
	cout << "\t" << (*vit);
      cout << flush << endl;
      #endif
      
      // step 1: copy the expression profile with index indormation
      unsigned int index;
      for (index = 0; index < ePlength; index++) {
	idx2valueVect[index].first = index;                 // original index
	idx2valueVect[index].second = ePMit->second[index]; // value
      }

      // step 2: sort the profile according to decreasing expr. values
      sort(idx2valueVect.begin(), idx2valueVect.end(),
	   compareUnsignedInt2DoubleDesc);

      // step 3: replace the original profile by ranks
      unsigned int sumRank = 0;
      unsigned int numRanksWithValue = 0;
      double lastValue = NO_EXPRVALUE;
      vector<double> origIndicesWithValue;

      unsigned int thisRank = 0;
      vector<pair<unsigned int, double> >::iterator i2vIt;
      for (i2vIt = idx2valueVect.begin(); i2vIt != idx2valueVect.end();
	   i2vIt++) {

	thisRank++;
	unsigned int orig_index = i2vIt->first;
	double thisValue = i2vIt->second;

	if (thisValue == lastValue) {
	  sumRank += thisRank;  // same value, need to average over the ranks
	  numRanksWithValue++;  // later!
	  origIndicesWithValue.push_back(orig_index);

          //#ifdef DEBUG
	  //cout << "#DEBUG: SAME VALUE! thisRank=" << thisRank << "; orig_index="
	  //     << orig_index << "; thisValue=" << thisValue
	  //     << "; lastValue="<< lastValue << "; sumRank=" << sumRank
	  //     << "; numRanksWithValue=" << numRanksWithValue << flush << endl;
          //#endif
	}
	else {
	  // write the previous average (if we had more data points with the
	  // same value)
	  vector<double>::iterator oIit;
	  for (oIit = origIndicesWithValue.begin();
	       oIit != origIndicesWithValue.end(); oIit++) {
	    ePMit->second[*oIit] = (double)sumRank/(double)numRanksWithValue;
	  }
	  origIndicesWithValue.clear();

	  // this is a new value
	  sumRank = thisRank;
	  numRanksWithValue = 1;

	  //#ifdef DEBUG
	  //cout << "#DEBUG: NEW VALUE! thisRank=" << thisRank << "; orig_index="
	  //     << orig_index << "; thisValue=" << thisValue
	  //     << "; lastValue="<< lastValue << "; sumRank=" << sumRank
	  //     << "; numRanksWithValue=" << numRanksWithValue << flush << endl;
          //#endif

	  lastValue = thisValue;
	  origIndicesWithValue.push_back(orig_index);
	}


      } // for (index = 0; ...

      // write also the last!
      vector<double>::iterator oIit;
      for (oIit = origIndicesWithValue.begin();
	   oIit != origIndicesWithValue.end(); oIit++) {
	ePMit->second[*oIit] = (double)sumRank/(double)numRanksWithValue;
      }
      origIndicesWithValue.clear();


      #ifdef DEBUG
      cout << "#DEBUG:" << endl;
      cout << "#DEBUG:     ranked:";
      for (vector<double>::iterator vit = ePMit->second.begin();
	   vit != ePMit->second.end(); vit++)
	cout << "\t" << (*vit);
      cout << endl;
      cout << "#DEBUG:" << endl;
      #endif

    } // for ( ; ePMit ...



  } // else if (coExprScoringFunction == "spearman") ...



  if (useDGPprob) {
    // read DGP data dile:
    #ifdef DEBUG
    cout << "#DEBUG: [========== File: DGP data ==========] " << endl;
    #endif

    if ( (exitStatus=parseDGPDataFile(dgpFileName, geneDGPmap)) != 0) {
      cerr << "Error: Couldn't read file " << dgpFileName
	   << " ... quitting! Errorcode: " << exitStatus << endl;
      exit(exitStatus);
    }
  }



  // loop over all disease genes!
  #ifdef DEBUG
  cout << "#DEBUG: [========== Loop over disease gene-OMIM pairs ==========] "
       << endl;
  #endif

  set<string>::iterator disGit;

  for (disGit = disGenes4L1O.begin(); disGit != disGenes4L1O.end(); disGit++) {

    vector<string> refGenes;
    vector<string> candGenes;

    map<string,map<string,double> > coexprListsScoreMap;
                               // maps coexpr. lists to a gene-CoexScore map
    map<string,map<string,double> > coexprListsRankMap;
                               // maps coexpr. lists to a gene-(rel.)rank map

    vector<string> rankedCandGenes;           // final ranked list of candidates.
    map<string,double> candGene2ScoreMap;     // to store overall scores.
    map<string,double> candGene2AvCoexScoreMap; // to store average CoexScores.

    map<string,vector<double> > candGene2relPositionsMap; // for keeping track
                                          // of the relative positions the
                                          // candidates had in the reference
                                          // gene co-expression lists or vice
                                          // versa; for extended output.

    vector<string> disGTokens;
    if (splitString(*disGit, '\t', disGTokens) != 0) {
      #ifdef DEBUG
      cout << "#DEBUG: Error while parsing disease gene-OMIM pair '" << *disGit
	   << "': cannot split string!" << endl;
      #endif

      return 129;
    }
    if (disGTokens.size() != 2) {

      #ifdef DEBUG
      cout << "#DEBUG: Error while parsing line: Wrong number of fields! "
	   << "Found: " << disGTokens.size() << ", expected: 2."
	   << endl;
      #endif

      return 130;
    }

    string thisDisGene= disGTokens[0];
    string thisDisPhenotype = disGTokens[1];
    #ifdef DEBUG
    cout << "#DEBUG: [========== Processing: disease gene " << thisDisGene
	 << " for OMIM " << thisDisPhenotype << " ==========] " << endl;
    #endif



    // get list of reference genes!
    #ifdef DEBUG
    cout << "#DEBUG: Reference genes";
    #endif

    /*
    // read input file for reference genes:
    #ifdef DEBUG
    cout << "#DEBUG: [========== File: reference genes ==========] " << endl;
    #endif

    if ( (exitStatus=parseRefGeneFile(refGeneFileName, refGenes)) != 0) {
      cerr << "Error: Couldn't read file " << refGeneFileName
	   << " ... quitting! Errorcode: " << exitStatus << endl;
      exit(exitStatus);
    }
    */

    set<string> rGUnique;


    map<string, set<string> >::iterator thisOmim2GenesIt =
                                         omim2GenesMap.find(thisDisPhenotype);

    // loop over similar phenotypes
    map<string, set<string> >::iterator omim2SimIt =
      omim2SimPhenoMap.find(thisDisPhenotype);

    if (omim2SimIt != omim2SimPhenoMap.end()) {

      set<string>::iterator simIt;
      for (simIt = (*omim2SimIt).second.begin();
	   simIt != (*omim2SimIt).second.end();
	   simIt++) {

	// for the similar phenotype, get involved genes:
	map<string, set<string> >::iterator sim2GenesIt
	                                    = omim2GenesMap.find(*simIt);

	if (sim2GenesIt != omim2GenesMap.end()) {

	  set<string>::iterator geneIt;

	  for (geneIt = (*sim2GenesIt).second.begin();
	       geneIt != (*sim2GenesIt).second.end();
	       geneIt++) {

	    // skip all genes that are also involved in the query phenotype?
	    if (removePhenoAnnoFromAllGenes
		&& thisOmim2GenesIt != omim2GenesMap.end()) {

	      if ( (*thisOmim2GenesIt).second.find(*geneIt)
		                  != (*thisOmim2GenesIt).second.end()) {
		// this reference gene is also associated to the query OMIM ID
		// so don't consider it (pretend unknown molecular basis)
		continue;
	      }
	    }

	    // otherwise remove at least THIS disease gene (leave-one-out)
	    if (*geneIt == thisDisGene) {
	      continue;
	    }


	    if (rGUnique.insert(*geneIt).second == true) { // new in list

	      refGenes.push_back(*geneIt);

	      #ifdef DEBUG
	      cout << ", " << *geneIt;
	      #endif
	    }

	  } // for (geneIt ...

	}

      } // for (simIt ...

    }

    #ifdef DEBUG
    cout << endl;
    #endif

    // do we have reference genes?
    if (refGenes.size() == 0) {
      #ifdef DEBUG
      cout << "#DEBUG: No reference genes, skipping!" << endl;
      #endif

      continue;
    }



    // get list of candidate genes!

    /*
    // read input file for candidate genes:
    #ifdef DEBUG
    cout << "#DEBUG: [========== File: candidate genes ==========] " << endl;
    #endif

    if ( (exitStatus=parseCandGeneFile(candGeneFileName, candGenes)) != 0) {
      cerr << "Error: Couldn't read file " << candGeneFileName
	   << " ... quitting! Errorcode: " << exitStatus << endl;
      exit(exitStatus);
    }
    */

    // NOT NEEDED ANYMORE
    /*
    set<string> cGUnique;

    string cmd = "rmp-getENSGrange.pl -i " + ensemblGeneListFile
      + " -g " + thisDisGene + " -n " + int2string(numAdjacentGenes)
      + " | extractSpecColumns.pl -c 1";


    #ifdef DEBUG
    cout << "#DEBUG: Executing: " << cmd << endl;
    cout << "#DEBUG: Candidate genes";
    #endif


    string candGList;
    int errcode = 0;
    if( (errcode = getStdOutFromCommand(cmd, candGList)) != 0) {
      cerr << "Error code " << errcode << " while executing: " << cmd << endl;
      exit(177);
    }

    vector<string> candGTokens;
    if (splitString(candGList, '\n', candGTokens) != 0) {
      cerr << "Error while parsing stdout of: " << cmd << endl
	   << "had: <STDOUT>" << candGList << "</STDOUT>" << endl;
      exit (188);
    }

    vector<string>::iterator cGTit;
    for (cGTit = candGTokens.begin(); cGTit != candGTokens.end(); cGTit++) {
      if (*cGTit != "" && cGUnique.insert(*cGTit).second == true) { // not empty
	                                                      // and new in list

	  candGenes.push_back(*cGTit);

	  #ifdef DEBUG
	  cout << ", " << *cGTit;
	  #endif
      }
    }

    #ifdef DEBUG
    cout << endl;
    #endif
    */


    int errcode = getCandGenesFromChrGeneList(chrGeneList, gene2ListPos,
					      thisDisGene, numAdjacentGenes,
					      candGenes);
    if (errcode != 0) {
      cerr << "Error code " << errcode
	   << " while determining list of candidate genes!" << endl;
      exit(errcode);
    }

    // do we have candidates?
    if (candGenes.size() == 0) {
      #ifdef DEBUG
      cout << "#DEBUG: No candidates, skipping!" << endl;
      #endif

      continue;
    }


    // for each reference gene, initialize an unordered coexpr. list:
    unsigned int g;
    vector<string> *exprListBaseGenes = &refGenes;
    if (coexprLists4CandGenes) {
      exprListBaseGenes = &candGenes;
    }

    for (g = 0; g < (*exprListBaseGenes).size(); g++) {
      map<string,double> newGene2CoexScoreMap;
      pair<string,map<string,double> > newPair;
      newPair.first = (*exprListBaseGenes)[g];    // base gene ID (cand or ref)
      newPair.second = newGene2CoexScoreMap;      // map for genes and coexScores

      bool inserted = (coexprListsScoreMap.insert(newPair)).second;

      if (!inserted) {
	cerr << "Error: could not create map for co-expression scores of co-expression list base gene " << (*exprListBaseGenes)[g] << " ... quitting!" << endl << flush;
	exit(111);
      }
    }


    // compute co-expression scores and build coexpr. lists for base genes
    #ifdef DEBUG
    cout << "#DEBUG: [========== Build coexpression lists ==========] " << endl;
    #endif

    if ( (exitStatus=buildCoexprLists4BaseGenes(exprProfileMap,
						computeCoExprTwoGenes,
						coexprListsScoreMap,
						weights, functSpecInfo)) != 0) {
      cerr << "Error: Couldn't build coexpression lists ... quitting! "
	   << "Errorcode: " << exitStatus << endl;
      exit(exitStatus);
    }


    // now start sorting the co-expression lists:
    #ifdef DEBUG
    cout << "#DEBUG: [========== Sorting co-expression lists ==========] " << endl;
    #endif

    if ( (exitStatus=sortCoexprLists(coexprListsScoreMap,
				     coexprListsRankMap)) != 0) {
      cerr << "Error: Couldn't correctly sort the co-expression lists"
	   << " ... quitting! Errorcode: " << exitStatus << endl;
      exit(exitStatus);
    }

    // we have the ranks of all genes in the ce-expression lists, now we can
    // compute scores for the candidate genes!
    #ifdef DEBUG
    cout << "#DEBUG: [========== Computing overall scores for candidates ==========] " << endl;
    #endif


    // to count number of candidates that could be scored
    int numCandGenesPresent = 0;


    for (g = 0; g < candGenes.size(); g++) {
      // iteration over all candidate genes

      #ifdef DEBUG
      cout << "#DEBUG: Computing score for candidate " << candGenes[g] << endl;
      #endif

      // compute score and average relative rank:
      double thisScore = UNSCORED;

      double sumRelRank = 0.0;
      int numRelRanks = 0;

      // either overall rank prob of this candidate on all reference coexpression
      // lists (default); or rank prob of all reference genes agains this
      // candidate's co-expression list (option -C on command line).
      vector<string> searchGenes;
      vector<string> coexprListGenes;
      if (!coexprLists4CandGenes) { // default
	searchGenes.push_back(candGenes[g]);
	coexprListGenes = refGenes;
      }
      else { // option -C
	searchGenes = refGenes;
	coexprListGenes.push_back(candGenes[g]);
      }

      // now get scores:
      vector<string>::iterator sGit;
      vector<string>::iterator cLGit;
      for (sGit = searchGenes.begin(); sGit != searchGenes.end(); sGit++) {
	// for each search gene (either this candidate or all reference genes):

	for (cLGit = coexprListGenes.begin(); cLGit != coexprListGenes.end();
	     cLGit++) {
	  // for each co-expression list (either this gene's or those of all
	  // reference genes):

	  // find rank if we have one (maybe not all genes have expression data):
	  map<string,map<string,double> >::iterator ceLRMit =
	    coexprListsRankMap.find(*cLGit);

	  if (ceLRMit != coexprListsRankMap.end()) {
	    // we have a co-expression list (might also be empty).
	    map<string,double>::iterator rankMapIt = 
	      ((*ceLRMit).second).find(*sGit);

	    if (rankMapIt != ((*ceLRMit).second).end()) {
	      // we have a relative rank for this search gene in the coexpr.list!
	      double thisRank = (*rankMapIt).second;

	      if (thisScore == UNSCORED)
		thisScore = 1; // multipying this with all relative ranks gives
	                       // an overall probability of having such ranks
	                       // by chance!
	      thisScore *= thisRank;

	      sumRelRank += thisRank;
	      numRelRanks++;

	      // keep track of single relative positions, if extended output:
	      if (extendedOutput) {
		string cG;
		if (!coexprLists4CandGenes) { // default
		  cG = *sGit;
		}
		else {
		  cG = *cLGit;
		}

		map<string,vector<double> >::iterator rPit = 
		  candGene2relPositionsMap.find(cG);

		// is this candidate gene already in the map?
		if (rPit == candGene2relPositionsMap.end()) { // not yet in map

		  vector<double> newVec;
		  pair<string,vector<double> > newPair;
		  newPair.first = cG;
		  newPair.second = newVec;

		  if ((candGene2relPositionsMap.insert(newPair)).second !=true) {
		    cerr << "Error: couldn't add " << newPair.first
			 << " to map of relative positions in co-expression "
			 << "lists!" << endl;
		    exit(55);
		  }

		  rPit = candGene2relPositionsMap.find(cG);
		}

		// add the relative position to the vector!
		(*rPit).second.push_back(thisRank);

		//#ifdef DEBUG
		//cout << "#DEBUG: added relative position (" << thisRank
		//   << ") for search gene " << (*sGit) << " in co-expression "
		//   << "list of gene " << (*cLGit) << " to relative position "
		//   << "map of candidate gene " << cG << endl;
		//#endif

	      } // if (extendedOutput) ...


              #ifdef DEBUG
	      cout << "#DEBUG: Search gene " << (*sGit) << " has relative rank "
		   << thisRank
		   << " in co-expression list of base gene " << (*cLGit)
		   << " [score=" << thisScore << "]" << endl;
              #endif
	    }
	    else {
	      #ifdef DEBUG
	      cout << "#DEBUG: No relative rank for search gene " << (*sGit)
		   << " in co-expression list of base gene " << (*cLGit)
		   << "!" << endl;
              #endif
	    }
	  }
	  else {
	    #ifdef DEBUG
	    cout << "#DEBUG: No co-expression list for base gene " << (*cLGit)
		 << "!" << endl;
            #endif
	  }

	} // for (cLGit = ...

      } // for (sGit = ...

      #ifdef DEBUG
      cout << "#DEBUG: Overall score for candidate gene " << candGenes[g]
	   << ": ";
      if (thisScore == UNSCORED)
	cout << "UNSCORED";
      else
	cout << thisScore;
      cout << endl << flush;
      #endif


      if (useDGPprob) {
	map<string,double>::iterator gDGPit = geneDGPmap.find(candGenes[g]);
	if (gDGPit == geneDGPmap.end()) {
	  #ifdef DEBUG
	  cout << "#DEBUG: No DGP prob. for gene " << candGenes[g]
	       << "! Considering as unscored!" << endl;
          #endif

	  thisScore = UNSCORED;
	}
	else { // got a DGP prob.
	  double dgp = (*gDGPit).second;
	  double overallScore =  thisScore * (1.0 - dgp);
	                          // multiply with score from co-expr. lists,
	                          // but keep in mind that higher DGP prob.
	                          // = more likely to be a disease gene (while
	                          // we have lower scores for better candidates

	  #ifdef DEBUG
	  cout << "#DEBUG: DGP prob. for gene " << candGenes[g]
	       << ": " << dgp << " => overall score: " << thisScore 
	       << " * (1.0 - " << dgp << ") = " << overallScore << endl;
          #endif

	  thisScore = overallScore;
	}

      }


      // count genes that could have been scored!
      if (thisScore != UNSCORED)
	numCandGenesPresent++;


      // insert into score map!
      pair<string,double> newPair;
      newPair.first = candGenes[g];
      newPair.second = thisScore;

      if ( (candGene2ScoreMap.insert(newPair)).second != true) {
	cerr << "Error: couldn't insert overall score of candidate "
	     << candGenes[g] << " into score map! Quitting!" << endl;
	exit(56);
      }

      // compute average relative rank!
      double avRelRank = UNSCORED;
      if (numRelRanks != 0) {
	avRelRank = sumRelRank / (double) numRelRanks;

        #ifdef DEBUG
	cout << "#DEBUG: Average relative rank of candidate gene "
	     << candGenes[g]
	     << ": " // << sumRelRank << " / " << numRelRanks << " = "
	     << avRelRank << endl << flush;
        #endif

	// insert into map:
	pair<string,double> newPair;
	newPair.first = candGenes[g];
	newPair.second = avRelRank;

	if ( (candGene2AvCoexScoreMap.insert(newPair)).second != true) {
	  cerr << "Error: couldn't insert average relative rank of candidate " 
	       << candGenes[g] << " into score map! Quitting!" << endl;
	  exit(57);
	}

      }

    } // for (g = ...


    // now rank candidates:
    #ifdef DEBUG
    cout << "#DEBUG: [========== Ranking candidates ==========] " << endl;
    #endif

    if ( (exitStatus=rankCandidateGenes(candGene2ScoreMap,
					rankedCandGenes)) != 0) {
      cerr << "Error: Couldn't correctly rank candidate genes"
	   << " ... quitting! Errorcode: " << exitStatus << endl;
      exit(exitStatus);
    }


    // produce final output:
    #ifdef DEBUG
    cout << "#DEBUG: [========== Produce final ouptut ==========] " << endl;
    #endif

    // DO SOME STATISTICS:
    // - fraction of reference genes in expression data?
    // - fraction of candidate genes in expression data?
    // - average pairwise correlation score of reference genes?
    unsigned int numRefGenesPresent = 0;
    unsigned int numRefGenesPresent4ThisOMIM = 0;
    unsigned int numRefGenes4ThisOMIM = 0; 
    vector<string> presentRefGenes;

    unsigned int rG;
    for (rG = 0; rG < refGenes.size(); rG++) {

      bool geneAssoc2ThisOMIM = false;
      if (omim2GenesMap.find(thisDisPhenotype) != omim2GenesMap.end()) {
	set<string>::iterator g_it =
	    (*omim2GenesMap.find(thisDisPhenotype)).second.find(refGenes[rG]);

	if (g_it != (*omim2GenesMap.find(thisDisPhenotype)).second.end()) {
	  geneAssoc2ThisOMIM = true;
	  numRefGenes4ThisOMIM++;
	}
      }

      if (genesWithExprData.find(refGenes[rG]) != genesWithExprData.end()) {
	numRefGenesPresent++;
	presentRefGenes.push_back(refGenes[rG]);

	if (geneAssoc2ThisOMIM) {
	  numRefGenesPresent4ThisOMIM++;
	}
      }
    }
  
    double fracRefGenesWithExprData =
      (double)numRefGenesPresent / (double)refGenes.size();

    // ALREADY COUNTED WHILE SCORING
    //int numCandGenesPresent = 0;
    //unsigned int cG;
    //for (cG = 0; cG < candGenes.size(); cG++) {
    //  if (genesWithExprData.find(candGenes[cG]) != genesWithExprData.end())
    //    numCandGenesPresent++;
    //}

    double fracCandGenesWithExprData =
      (double)numCandGenesPresent / (double)candGenes.size();

    double avRefCorrScore = computeAveragePairwiseScore(presentRefGenes,
							exprProfileMap,
							computeCoExprTwoGenes,
							weights, functSpecInfo);

    cout << "+++ OMIM ID=" << thisDisPhenotype << " gene=" << thisDisGene
	 << " +++" << endl;
    cout << "--- OMIM ID=" << thisDisPhenotype << " gene=" << thisDisGene
	 << " ---" << endl;
    cout << "# Number of reference genes with expression profile associated to "
	 << "OMIM ID " << thisDisPhenotype << ": "
	 << numRefGenesPresent4ThisOMIM << " (of " << numRefGenes4ThisOMIM
	 << " associated genes)" << endl;
    cout << "# mbc-evalCandGenesByCoExpr" << endl;
    cout << "# Version " << PRGVERSION << endl;
    cout << "# Command: ";
    for (int jj = 0; jj < argc; jj++) {
      cout << argv[jj] << " ";
    }
    cout << endl;
    cout << "# Date: " << dateString;
    cout << "# Reference genes for OMIM(+similar): " << thisDisPhenotype << endl;
    cout << "# Number of reference genes: " << refGenes.size() << endl;

    cout << "# Number of reference genes having expression data: "
	 << numRefGenesPresent << " (fraction="
	 << fracRefGenesWithExprData << ")" << endl << flush;
    cout << "# Average pairwise correlation score of reference genes: ";
    if (avRefCorrScore == DBL_MIN) {
      cout << "-";
    }
    else {
      cout << avRefCorrScore;
    }
    cout << endl << flush;

    cout << "# Candidate genes for artificial locus of: " << thisDisGene
	 << " (+/- " << int2string(numAdjacentGenes) << ")" << endl;
    cout << "# Number of candidate genes: " << candGenes.size() << endl;

    cout << "# Number of candidate genes having necessary data for scoring (expression, DGP): "
	 << numCandGenesPresent << " (fraction="
	 << fracCandGenesWithExprData << ")" << endl << flush;

    cout << "# Expression data file: " << exprDataFileName << endl;
    cout << "# Scoring function for co-expression: " << coExprScoringFunction
	 << endl;
    if (minNormExpr != DBL_MIN) {
      cout << "# Minimum normalized expression: " << minNormExpr
	   << endl;
    }
    if (weightsFileName != "") {
      cout << "# Weight vector for expression profiles: " << weightsFileName
	   << endl;
    }

    cout << "# Sorted co-expression lists built for: ";
    if (coexprLists4CandGenes)
      cout << "candidate genes";
    else
      cout << "reference genes";
    cout << endl;

    cout << "# Using DGP probabilities: ";
    if (useDGPprob)
      cout << "yes (file: " << dgpFileName << ")";
    else
      cout << "no";
    cout << endl;


    // list of prioritized candidates:

    // header:
    cout << "rank\t"
	 << "gene\t"
	 << "score_mult\t"
	 << "score_av\t"
	 << "relRank_scored\t"
	 << "relRank_all";
    if (useDGPprob)
      cout << "\tscore_dgp";

    if (extendedOutput) {
      for (unsigned int jj = 0; jj < numRefGenesPresent; jj++) {
	cout << "\trelPos_" << presentRefGenes[jj];
      }
    }
    cout << endl;

    for (unsigned int candRank = 1; candRank <= rankedCandGenes.size();
	 candRank++) {
      // - average Pearson for each candidate (either candidates or ref. genes in
      //   coexpression lists)
      // - relative rank of candidates regarding ALL candidates and regarding
      //   only scored candidates

      string thisCand = rankedCandGenes[candRank-1];
      double thisScore = (*(candGene2ScoreMap.find(thisCand))).second;

      double thisAvCoexScore =
	                 (*(candGene2AvCoexScoreMap.find(thisCand))).second;
      double relRankAmongScoredCands = (double)candRank / 
	                                       (double)numCandGenesPresent;
      double relRankAmongAllCands = (double)candRank / 
	                                    (double)candGenes.size();

      cout << candRank << "\t"
	   << thisCand << "\t";

      if (thisScore == UNSCORED) {
	cout << "unscored\t"
	     << "unscored\t"
	     << "-" << "\t";
      }
      else {
	cout << thisScore << "\t"
	     << thisAvCoexScore << "\t"
	     << relRankAmongScoredCands << "\t";
      }

      cout << relRankAmongAllCands;

      if (useDGPprob) {
	map<string,double>::iterator gDGPit = geneDGPmap.find(thisCand);
	if (gDGPit == geneDGPmap.end())
	  cout << "\tunscored";
	else
	  cout << "\t" << (*gDGPit).second;
      }

      if (extendedOutput) {

	if (thisScore == UNSCORED) {
	  for (unsigned int jj = 0; jj < numRefGenesPresent; jj++) {
	    cout << "\t-";
	  }
	}
	else {
	  map<string,vector<double> >::iterator rPit =
	    candGene2relPositionsMap.find(thisCand); // must exist!

	  for (unsigned int jj = 0; jj < numRefGenesPresent; jj++) {
	    cout << "\t" << (*rPit).second[jj];
	  }
	}
      }


      cout << endl;
    }


  } // loop over disease gene list for leave-one-out!




  exit(exitStatus);

} // main() ...



//----------------------------------------------------------------------


//--------------------------------------/
// --- getCandGenesFromChrGeneList() ---/
//--------------------------------------/

int getCandGenesFromChrGeneList(vector<pair<string, string> > &chrGeneList,
				map<string,unsigned int> &gene2ListPos,
				const string thisDisGene,
				const unsigned int numAdjacentGenes,
				vector<string> &candGenes) {
  int retVal = 0;

  set<string> uniqueCandGenes;

  // first determine what position the true disease gene has in the ordered
  // vector:
  unsigned int disGenePos;
  string disGeneChromosome = "";

  map<string,unsigned int>::iterator g2LP_it = 
    gene2ListPos.find(thisDisGene);

  #ifdef DEBUG
  cout << "#DEBUG: Building artifical locus for candidate genes centered around " << thisDisGene << endl << flush;
  #endif


  if (g2LP_it == gene2ListPos.end()) {
    // the disease gene isn't in out gene list, this is OK (will be skipped),
    // so don't return an error!!!
    #ifdef DEBUG
    cout << "#DEBUG: Disease gene " << thisDisGene << " not found in the "
	 << "gene list file. Cannot create artificial locus." << endl << flush;
    #endif
      
    return 0;
  }

  disGenePos = (*g2LP_it).second;


  // now determine on which chromosome it resides:
  disGeneChromosome = chrGeneList[disGenePos].first;

  #ifdef DEBUG
  cout << "#DEBUG: " << thisDisGene << " is on chromosome "
       << disGeneChromosome << " and has position " << disGenePos
       << " in the Ensembl gene list." << endl << flush;
  #endif


  if (thisDisGene != chrGeneList[disGenePos].second) { // do a fast check
    // something wrong, seems to be a different gene, that shouldn't happen
    // if parseGeneListFile() reads the file correctly.
    #ifdef DEBUG
    cout << "#DEBUG: Error. Expected gene " << thisDisGene
	 << " at position " << disGenePos << " in chrGeneList!" << endl
	 << flush;
    #endif
    retVal = 237;
  }

  if (disGeneChromosome == "") {
    #ifdef DEBUG
    cout << "#DEBUG: Error. No chromosome defined for gene " << thisDisGene
	 << " at position " << disGenePos << " in chrGeneList!" << endl
	 << flush;
    #endif
    retVal = 238;
  }


  // now check the necessary adjacent genes and add them to the artificial
  // locus under two conditions: 1) same chromosome; 2) not yet added
  unsigned int lowerPos = 0;
  unsigned int upperPos = chrGeneList.size()-1;

  if ((int)disGenePos - (int)numAdjacentGenes > (int)lowerPos)
    lowerPos = disGenePos - numAdjacentGenes;

  if (disGenePos + numAdjacentGenes < upperPos)
    upperPos = disGenePos + numAdjacentGenes;

  #ifdef DEBUG
  cout << "#DEBUG: Checking Ensembl gene file range: " << lowerPos
       << " to " << upperPos << endl << flush;
  #endif


  unsigned int ii;
  for (ii = lowerPos; ii <= upperPos; ii++) {
    if ( (chrGeneList[ii].first == disGeneChromosome)
	 && (uniqueCandGenes.find(chrGeneList[ii].second) ==
	     uniqueCandGenes.end())
	 ) {

      // add to artificial locus:
      candGenes.push_back(chrGeneList[ii].second);

      // make sure we add it only once:
      uniqueCandGenes.insert(chrGeneList[ii].second);

      #ifdef DEBUG
      cout << "#DEBUG: Gene " << chrGeneList[ii].second << " at pos. "
	   << ii << " is on chromosome " << chrGeneList[ii].first 
	   << " ... adding to artifical locus!" << endl << flush;
      #endif
    }
    else {
      #ifdef DEBUG
      cout << "#DEBUG: Gene " << chrGeneList[ii].second << " at pos. "
	   << ii << " is on chromosome " << chrGeneList[ii].first 
	   << " ... excluding from artificial locus!" << endl << flush;
      #endif
    }

  }

  return retVal; // ok?

} // getCandGenesFromChrGeneList() ...


//------------------------/
//--- printUsageInfo() ---/
//------------------------/

void printUsageInfo(const char *prgname) {
  cerr << "Author: Rosario M. Piro (piro@to.infn.it)" << endl
       << "Usage: " << prgname << " [OPTIONS]" << endl << endl
       << "where options are:" << endl
       << " -h                Print this help message and quit." << endl
       << " -g <genelistfile> The name of the file that contains the BioMart output with" << endl
       << "                   the Ensembl gene list, that is already sorted according" << endl
       << "                   to the gene locus." << endl
       << "                   Important: The first line must contain the header with" << endl
       << "                   at least the following entries: \"" << GENEID_COLUMN_NAME << "\" and"<< endl
       << "                   \"" << CHR_COLUMN_NAME << "\". Genes must already be ordered" << endl
       << "                   according to their position on the chromosome." << endl
       << " -n <num>          Number of adjacent genes of known disease gene to be" << endl
       << "                   considered as candidate genes. Note: <num> genes from" << endl
       << "                   both directions on the chromosome will be taken, hence the" << endl
       << "                   total number of candidates will be 2x<num> + 1. Default: " << DEF_NUM_ADJ_GENES << endl
       << " -p <relgenefile>  The file containing the list of all genes associated to query" << endl
       << "                   and related phenotypes. The file should have the same format" << endl
       << "                   as the output of \"rmp-findDiseaseRelatedGenes.pl -O all\"." << endl
       << " -m <e2omapfile>   The file containing a mapping from gene to OMIM IDs." << endl
       << "                   This is the list of genes/phenotypes for the leave-one-out." << endl
       << " -R                Strip the OMIM phenotype annotation (regarding the query" << endl
       << "                   disease) not only from the query gene (from <e2omapfile>)," << endl
       << "                   but from all genes (in order to simulate a phenotype with" << endl
       << "                   unknown molecular basis)." << endl
       << " -e <file>         File containing the expression profiles; one profile per line" << endl
       << "                   Format: \"<gene>\\t<exprVal1>\\t<exprVal2>...\""<< endl
       << " -w <file>         File containing a list of weights of the single experiments/data" << endl
       << "                   points of the expression vectors; can be used for the scoring" << endl
       << "                   functions: pearson, quadraticpearson, euclidean, cosine," << endl
       << "                   spearman, relintoverlap, pearson3d (see below)." << endl
       << "                   If not specified, no weighting will be used." << endl
       << " -s <func>         Scoring function to use for computing the co-expression of" << endl
       << "                   two genes. Can be one of:" << endl
       << "                   \"pearson\": Pearson correlation coefficient r (default)" << endl
       << "                   \"quadraticpearson\": r*r (emphasizes higher values and" << endl
       << "                                            considers also anti-correlation)" << endl
       << "                   \"spearman\": Spearman rank correlation" << endl
       << "                   \"euclidean\": (inverse) Euclidean distance" << endl
       << "                   \"cosine\": Cosine similarity" << endl    
       << "                   \"relintoverlap\": Relative intensity overlap" << endl
       << "                   \"pearson3d\": average of Pearson correlation coefficients" << endl
       << "                          computed for every plane of the 3D space. This" << endl
       << "                          requires an expression profile with positional" << endl
       << "                          information in the header; for each column: \"X,Y,Z\"." << endl
       << " -M <double>       Minimum expression value." << endl
       << "                   All expression values below this threshold will be set to 0." << endl
       << "                   This is currently implemented only for the scoring function" << endl
       << "                   \"relintoverlap\" and will be ignored otherwise." << endl
       << " -C                Build co-expression lists for the candidate genes instead of" << endl
       << "                   the reference genes (check ranks of reference genes within" << endl
       << "                   the co-expression lists of candidate genes instead of checking" << endl
       << "                   the ranks of candidate genes within the reference genes' co-" << endl
       << "                   expression lists)." << endl
       << " -d <file>         Integrate DGP (Lopez-Bigas et al.) disease gene probabilities" << endl
       << "                   taken from the file. If not specified, DGP probilities are" << endl
       << "                   not considered." << endl
       << " -E                Extended output (relative positions in co-expression lists)." << endl << endl

       << "The options -g, -p, -m, and -e are mandatory." << endl << endl;
} // printUsageInfo() ...

