/*****************************************************************************
 * rmp-buildRCGsFromExpressionMatrix.cpp - Source file for building Ranked   *
 * Coexpression Groups (RCGs) from a gene expression matrix (see             *
 * Miozzi et al., PLoS One, 3(6), e2439, 2008).                              *
 * Version 0.0.1                                                             *
 * ------------------------------------------------------------------------- *
 * Copyright 2012 by Rosario Piro (mailto: rmpiro@gmail.com)                 *
 *                                                                           *
 *   This program is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation; _exclusively_ version 2 of the License    *
 *   NO later version.                                                       *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program; if not, write to the Free Software             *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA *
 *   or see http://www.gnu.org/copyleft/gpl.html                             *
 * ------------------------------------------------------------------------- */

//---------------------
//--- include files ---
//---------------------
//#include <fstream>
#include <iostream>
//#include <map>
//#include <set>
//#include <vector>
#include <iomanip>
//#include <cmath>
#include <cstring>
#include <algorithm>
//#include <cstdlib>
//#include <climits>
#include "extstdlib.h"
#include "evalCandGenesByCoExpr-serviceroutines.h"
//#include <ctime>


//---------------------------
//--- preprocessor macros ---
//---------------------------
#define PRGVERSION "0.0.1"

#define K_NOT_SPECIFIED 0

#define NO_EXPRVALUE DBL_MIN
#define NO_CORRVALUE DBL_MIN


//-----------------------------/
//--- function declarations ---/
//-----------------------------/

void printUsageInfo(const char *prgname);

//inline bool invDoubleComp(double i, double j) { return (i > j); }


//------------------------/
//--- global variables ---/
//------------------------/

//------------
//--- main ---
//------------

int main (int argc, const char **argv) {

  int exitStatus = 0;

  set<string> genesWithExprData;
  vector<string> headerInfo;
  map<string,vector<unsigned int> > planes; // needed only for pearson3d

  unsigned int rcgK = K_NOT_SPECIFIED; // number of genes per RCG apart from the center => size = 1+k

  map<string,vector<double> > exprProfileMap;

  map<string,vector<pair<string,double> > > rcgMap;

  vector<double> weights; // default: empty, no weights used!

  void *functSpecInfo = NULL;
  double minNormExpr = DBL_MIN; // used only for "relintoverlap"; here: always no

  double (*computeCoExprTwoGenes)(vector<double> &,
				  vector<double> &,
				  vector<double> &,
				  void *) = computePearson;
                               // points to the function that determins
                               // the co-expression of two genes.
                               // Default: Pearson correlation coefficient
  string coExprScoringFunction = "pearson";
  double coExprScoringFunctionPerfectCorr = 1.0; // perfect correlation

  bool maximizeScores = true; // default: higher similarity scores are better (e.g. Pearson)

  bool useAntiCorrelation = false; 

  string weightsFileName = "";
  string exprDataFileName = "";


  char *dateString;

  cout << setprecision(DEFAULT_PRECISION);

  // parse command line arguments:
  int index = 1;
  while (index < argc) {
    // help
    if (strcmp(argv[index],"--help")==0 || strcmp(argv[index],"-h")==0) {
      printUsageInfo(argv[0]);
      exit(0);
    }
    // Expression data file
    else if (strcmp(argv[index],"-e")==0) {
      if (index + 1 < argc) {
	exprDataFileName = argv[++index];
      }
      else { // filename missing after -e
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    // Weight file
    else if (strcmp(argv[index],"-w")==0) {
      if (index + 1 < argc) {
	weightsFileName = argv[++index];
      }
      else { // filename missing after -w
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    // coexpression score for two genes:
    else if (strcmp(argv[index],"-s")==0) {
      if (index + 1 < argc) {
	++index;
	if (strcmp(argv[index],"pearson")==0) {
	  computeCoExprTwoGenes = computePearson;
	  coExprScoringFunction = "pearson";
	  coExprScoringFunctionPerfectCorr = 1.0;
	  maximizeScores = true;
	}
	else if (strcmp(argv[index],"quadraticpearson")==0) {
	  computeCoExprTwoGenes = computeQuadraticPearson;
	  coExprScoringFunction = "quadraticpearson";
	  coExprScoringFunctionPerfectCorr = 1.0;
	  maximizeScores = true;
	}
	else if (strcmp(argv[index],"euclidean")==0) {
	  computeCoExprTwoGenes = computeEucDist;
	  coExprScoringFunction = "euclidean";
	  coExprScoringFunctionPerfectCorr = 0.0;
	  maximizeScores = false;
	}
	else if (strcmp(argv[index],"cosine")==0) {
	  computeCoExprTwoGenes = computeCosine;
	  coExprScoringFunction = "cosine";
	  coExprScoringFunctionPerfectCorr = 1.0;
	  maximizeScores = true;
	}
	else if (strcmp(argv[index],"spearman")==0) {
	  computeCoExprTwoGenes = computePearson;  // ! spearman = pearson on
	                                           //   rank transformed data!
	  coExprScoringFunction = "spearman";
	  coExprScoringFunctionPerfectCorr = 1.0;
	  maximizeScores = true;
	}
	else if (strcmp(argv[index],"relintoverlap")==0) {
	  computeCoExprTwoGenes = computeRelativeIntensityOverlap;
	  coExprScoringFunction = "relintoverlap";
	  coExprScoringFunctionPerfectCorr = 1.0;
	  maximizeScores = true;
	}
	//else if (strcmp(argv[index],"normpearson")) {
	//computeCoExprTwoGenes = computeNormalizedPearson;
	//coExprScoringFunction = "normpearson";
	//}
	else if (strcmp(argv[index],"pearson3d")==0) {
	  computeCoExprTwoGenes = computePearson3D;
	  coExprScoringFunction = "pearson3d";
	  coExprScoringFunctionPerfectCorr = 1.0;
	  maximizeScores = true;
	}
	else { // unknwon scoring function
	  printUsageInfo(argv[0]);
	  exit(1);
	}
      }
      else { // scoring function missing after -s
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    // RCG size -1 (i.e. number of partners)
    else if (strcmp(argv[index],"-k")==0) {
      if (index + 1 < argc) {
	const char *val = argv[++index];
	if (isNumeric(val))
	  rcgK = atoi(val);
	else {
	  printUsageInfo(argv[0]);
	  exit(1);
	}
      }
      else { // number missing after -k
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    // use anti-correlation instead?
    else if (strcmp(argv[index],"-A")==0) {
      useAntiCorrelation = true;
    }

    index++;
  }


  // check for mandatory parameters:
  if (exprDataFileName == "" || rcgK == K_NOT_SPECIFIED || rcgK < 0) {
    printUsageInfo(argv[0]);
    exit(1);
  }


  if (useAntiCorrelation) {
    // invert the choice of whether to maximiye scores
    maximizeScores = !maximizeScores;
  }


  // before starting determine date and time:
  time_t currTime = time(NULL);
  dateString = ctime(&currTime);


  // read input files

  if (weightsFileName != "") {
    // read input file for weight vector:
    #ifdef DEBUG
    cout << "#DEBUG: [========== File: weight vector ==========] " << endl;
    #endif

    if ( (exitStatus=parseWeightFile(weightsFileName, weights)) != 0) {
      cerr << "Error: Couldn't read file " << weightsFileName
	   << " ... quitting! Errorcode: " << exitStatus << endl;
      exit(exitStatus);
    }
  }


  // read expression data file (expression profiles):
  #ifdef DEBUG
  cout << "#DEBUG: [========== File: Expression data ==========] " << endl;
  #endif

  if ( (exitStatus=parseExpressionDataFile(exprDataFileName, genesWithExprData,
					   exprProfileMap, headerInfo)) != 0) {
    cerr << "Error: Couldn't read file " << exprDataFileName
	 << " ... quitting! Errorcode: " << exitStatus << endl;
    exit(exitStatus);
  }


  // set additional input information or perform additional pre-processing
  // for some of the correlation measures:

  // if needed, parse header info to extract planes:
  if (coExprScoringFunction == "pearson3d") {
    if ((exitStatus=convertHeaderInfo2Planes(headerInfo, planes)) != 0) {
      cerr << "Error: Couldn't convert header info to planes "
	   << " ... quitting! Errorcode: " << exitStatus << endl;
      exit(exitStatus);
    }

    functSpecInfo = &planes; // need this to pass
  }
  else if (coExprScoringFunction == "relintoverlap") {
    // if we don't need planes, we may need a minimum normalized expr.
    // (for noise reduction with relintoverlap)
    functSpecInfo = &minNormExpr;
  }
  else if (coExprScoringFunction == "spearman") {
    // Spearman rank correlation = Pearson correlation on rank transformed data
    #ifdef DEBUG
    cout << "#DEBUG: [========== Rank transformation of expression data for Spearman ==========] " << endl;
    #endif

    
    map<string,vector<double> >::iterator ePMit = exprProfileMap.begin();
    unsigned int ePlength = ePMit->second.size();

    vector<pair<unsigned int, double> > idx2valueVect;
    idx2valueVect.resize(ePlength);


    for ( ; ePMit != exprProfileMap.end(); ePMit++) {
      // for each expression profile ...

      #ifdef DEBUG
      cout << "#DEBUG: expression profile for " << ePMit->first << endl;
      cout << "#DEBUG:    profile:";
      for (vector<double>::iterator vit = ePMit->second.begin();
	   vit != ePMit->second.end(); vit++)
	cout << "\t" << (*vit);
      cout << flush << endl;
      #endif
      
      // step 1: copy the expression profile with index indormation
      unsigned int index;
      for (index = 0; index < ePlength; index++) {
	idx2valueVect[index].first = index;                 // original index
	idx2valueVect[index].second = ePMit->second[index]; // value
      }

      // step 2: sort the profile according to decreasing expr. values
      sort(idx2valueVect.begin(), idx2valueVect.end(),
	   compareUnsignedInt2DoubleDesc);

      // step 3: replace the original profile by ranks
      unsigned int sumRank = 0;
      unsigned int numRanksWithValue = 0;
      double lastValue = NO_EXPRVALUE;
      vector<double> origIndicesWithValue;

      unsigned int thisRank = 0;
      vector<pair<unsigned int, double> >::iterator i2vIt;
      for (i2vIt = idx2valueVect.begin(); i2vIt != idx2valueVect.end();
	   i2vIt++) {

	thisRank++;
	unsigned int orig_index = i2vIt->first;
	double thisValue = i2vIt->second;

	if (thisValue == lastValue) {
	  sumRank += thisRank;  // same value, need to average over the ranks
	  numRanksWithValue++;  // later!
	  origIndicesWithValue.push_back(orig_index);

          //#ifdef DEBUG
	  //cout << "#DEBUG: SAME VALUE! thisRank=" << thisRank << "; orig_index="
	  //     << orig_index << "; thisValue=" << thisValue
	  //     << "; lastValue="<< lastValue << "; sumRank=" << sumRank
	  //     << "; numRanksWithValue=" << numRanksWithValue << flush << endl;
          //#endif
	}
	else {
	  // write the previous average (if we had more data points with the
	  // same value)
	  vector<double>::iterator oIit;
	  for (oIit = origIndicesWithValue.begin();
	       oIit != origIndicesWithValue.end(); oIit++) {
	    ePMit->second[*oIit] = (double)sumRank/(double)numRanksWithValue;
	  }
	  origIndicesWithValue.clear();

	  // this is a new value
	  sumRank = thisRank;
	  numRanksWithValue = 1;

	  //#ifdef DEBUG
	  //cout << "#DEBUG: NEW VALUE! thisRank=" << thisRank << "; orig_index="
	  //     << orig_index << "; thisValue=" << thisValue
	  //     << "; lastValue="<< lastValue << "; sumRank=" << sumRank
	  //     << "; numRanksWithValue=" << numRanksWithValue << flush << endl;
          //#endif

	  lastValue = thisValue;
	  origIndicesWithValue.push_back(orig_index);
	}


      } // for (index = 0; ...

      // write also the last!
      vector<double>::iterator oIit;
      for (oIit = origIndicesWithValue.begin();
	   oIit != origIndicesWithValue.end(); oIit++) {
	ePMit->second[*oIit] = (double)sumRank/(double)numRanksWithValue;
      }
      origIndicesWithValue.clear();


      #ifdef DEBUG
      cout << "#DEBUG:" << endl;
      cout << "#DEBUG:     ranked:";
      for (vector<double>::iterator vit = ePMit->second.begin();
	   vit != ePMit->second.end(); vit++)
	cout << "\t" << (*vit);
      cout << endl;
      cout << "#DEBUG:" << endl;
      #endif

    } // for ( ; ePMit ...



  } // else if (coExprScoringFunction == "spearman") ...



  // start building the RCGs
  #ifdef DEBUG
  cout << "#DEBUG: [========== Building RCGs ==========] " << endl;
  #endif

  set<string>::iterator gWEDit[2];

  for (gWEDit[0] = genesWithExprData.begin(); gWEDit[0] != genesWithExprData.end(); gWEDit[0]++) {

    for (gWEDit[1] = gWEDit[0], gWEDit[1]++; gWEDit[1] != genesWithExprData.end(); gWEDit[1]++) {

      map<string, vector<pair<string,double> > >::iterator rcgMapIt;

      double corr = computeCoExprTwoGenes(exprProfileMap.find(*(gWEDit[0]))->second,
					  exprProfileMap.find(*(gWEDit[1]))->second,
					  weights, functSpecInfo);

      #ifdef DEBUG
      cout << "#DEBUG: Correlation(" << coExprScoringFunction << ") of "
	   << *(gWEDit[0]) << " and " << *(gWEDit[1]) << ": " << corr << endl;
      #endif


      // initialize RCGs for the two genes, if we don't have them yet
      for (unsigned int idx = 0; idx < 2; idx++) {
	if (rcgMap.find(*(gWEDit[idx])) == rcgMap.end()) {
	  vector<pair<string,double> > newRCGpartners;
	  pair<string, vector<pair<string,double> > > newPair;
	  newPair.first = *(gWEDit[idx]);
	  newPair.second = newRCGpartners;

	  if (!(rcgMap.insert(newPair)).second) {
	    cerr << "Error: could not create empty RCG for gene " << *(gWEDit[idx])
		 << " ... quitting!" << endl << flush;
	    exit(111);
	  }
	}
      }

      // see if the two genes are among the most correlated/anticorrelated of each other
      for (unsigned int idx = 0; idx < 2; idx++) {

        #ifdef DEBUG
	cout << "#DEBUG:    Checking RCG of " << *(gWEDit[idx]) << ": ";
        #endif

	rcgMapIt = rcgMap.find(*(gWEDit[idx]));

	bool insert = false; // determine whether we keep this corr for the RCG

	if (rcgMapIt->second.size() < rcgK) {
	  // RCG not yet at maximum size; ḱeep the corr in any case!
	  insert = true;

          #ifdef DEBUG
	  cout << "has " << rcgMapIt->second.size() << " additional members (<" << rcgK
	       << ") => inserting!" << endl << flush;
          #endif

	}
	else {
	  // RCG has already the maximum number of genes; check if this corr is needed
	  double lastRCGcorr = rcgMapIt->second.rbegin()->second;

	  if ( (maximizeScores && corr > lastRCGcorr)
	       || (!maximizeScores && corr < lastRCGcorr) ) {
	    // this correlation is better, need to keep it
	    insert = true;

            #ifdef DEBUG
	    cout << "last ranked member has corr=" << lastRCGcorr << " => inserting!"
		 << endl << flush;
            #endif
	  }
	  else {
            #ifdef DEBUG
	    cout << "last ranked member has corr=" << lastRCGcorr << " => ignoring this!"
		 << endl << flush;
            #endif
	  }
	}


	if (insert) {
	  // we need to insert it:
	  pair<string,double> newPair;
	  newPair.first = *(gWEDit[1-idx]);
	  newPair.second = corr;

	  rcgMapIt->second.push_back(newPair);

	  // sort the RCG
	  // NOTE: for large RCGs it would be better iterate through the already sorted RCG and
          // check where the new element has to be inserted; this could be done either from the
          // the beginning of the RCG or the end, depending on whether this elements correlation
          // is closer to the first or the last in the RCG. This should be faster than sorting
          // everything again.

	  if (maximizeScores) {
	    sort(rcgMapIt->second.begin(), rcgMapIt->second.end(), compareString2DoubleDesc);
	  }
	  else {
	    sort(rcgMapIt->second.begin(), rcgMapIt->second.end(), compareString2Double);
	  }

	  // do we now have too many in the RCG?
	  if (rcgMapIt->second.size() > rcgK) {
	    rcgMapIt->second.pop_back(); // delete last element
	  }

	} // if (insert) ...

	#ifdef DEBUG
	cout << "#DEBUG:             RCG now: " << *(gWEDit[idx]);
	vector<pair<string,double> >::iterator rcgIt;
	for (rcgIt = rcgMapIt->second.begin(); rcgIt != rcgMapIt->second.end(); rcgIt++) {
	  cout << ", " << rcgIt->first << "(" << rcgIt->second << ")"; 
	}
	cout << endl << flush;
        #endif

      } // for idx=0,1

      #ifdef DEBUG
      cout << "#DEBUG:" << endl << flush;
      #endif

    } // for (gWEDit2 ...

  } // for (gWEDit ...


  #ifdef DEBUG
  cout << "#DEBUG: [========== Produce final ouptut ==========] " << endl;
  #endif

  cout << "# rmp-buildRCGsFromExpressionMatrix" << endl;
  cout << "# Version " << PRGVERSION << endl;
  cout << "# Command: ";
  for (int jj = 0; jj < argc; jj++) {
    cout << argv[jj] << " ";
  }
  cout << endl;
  cout << "# Date: " << dateString;

  cout << "# Expression data file: " << exprDataFileName << endl;
  cout << "# Scoring function for co-expression: " << coExprScoringFunction << endl;
  if (weightsFileName != "") {
    cout << "# Weight vector for expression profiles: " << weightsFileName
	 << endl;
  }
  cout << "# Number of most ";
  if (useAntiCorrelation)
    cout << "anti-";
  cout << "correlated genes to add to each gene's RCG: " << rcgK << endl;


  // header:
  cout << "RCG\t"
       << "gene\t"
       << "rank\t"
       << coExprScoringFunction << "\t"
       << endl;

  set<string>::iterator git;
  for (git = genesWithExprData.begin(); git != genesWithExprData.end(); git++) {

    unsigned int rank = 1;

    // the center gene of this RCG itself
    cout << *git << "\t" << *git << "\t" << rank << "\t" << coExprScoringFunctionPerfectCorr << endl;

    vector<pair<string,double> >::iterator rcgIt;
    for (rcgIt = rcgMap.find(*git)->second.begin();
	 rcgIt != rcgMap.find(*git)->second.end(); rcgIt++) {

      cout << *git << "\t" << rcgIt->first << "\t" << (++rank) << "\t" << rcgIt->second << endl;
    }
  }

  exit(exitStatus);

} // main() ...



//----------------------------------------------------------------------

//------------------------/
//--- printUsageInfo() ---/
//------------------------/

void printUsageInfo(const char *prgname) {
  cerr << "Author: Rosario M. Piro (piro@to.infn.it)" << endl
       << "Usage: " << prgname << " [OPTIONS]" << endl << endl
       << "where options are:" << endl
       << " -h          Print this help message and quit." << endl
       << " -k <num>    Number of most co-expressed genes for RCGs; hence for each" << endl
       << "             each gene an RCG of size <num>+1 will be built."<< endl
       << "             (See Miozzi et al., PLoS One, 3(6), e2439, 2008)" << endl
       << " -e <file>   File containing the expression profiles (one profile per line; in" << endl
       << "             the form \"<gene>\\t<exprVal1>\\t<exprVal2>...\")."<< endl
       << " -w <file>   File containing a list of weights of the single experiments/data" << endl
       << "             points of the expression vectors; can be used for the scoring" << endl
       << "             functions: pearson, quadraticpearson, euclidean, cosine," << endl
       << "             spearman, relintoverlap, pearson3d (see below)." << endl
       << "             If not specified, no weighting will be used." << endl
       << " -s <func>   Scoring function to use for computing the co-expression of two" << endl
       << "             genes. Can be one of:" << endl
       << "             \"pearson\": Pearson correlation coefficient r (default)" << endl
    //<< "             \"normpearson\": normalized Pearson; for expression vectors with" << endl
    //<< "                            different numbers of expression values/experiments" << endl
       << "             \"quadraticpearson\": r*r (emphasizes higher values and considers" << endl
       << "                                 also anti-correlation)" << endl
       << "             \"spearman\": Spearman rank correlation" << endl
       << "             \"euclidean\": (inverse) Euclidean distance" << endl
       << "             \"cosine\": Cosine similarity" << endl    
       << "             \"relintoverlap\": Relative intensity overlap" << endl
       << "             \"pearson3d\": average of Pearson correlation coefficients" << endl
       << "                          computed for every plane of the 3D space. This" << endl
       << "                          requires an expression profile with positional" << endl
       << "                          information in the header; for each column: \"X,Y,Z\"." << endl
       << " -A          Find the <num> most anti-correlated genes for each RCG, instead" << endl
       << "             of the most correlated genes." << endl << endl
       << "The options -e and -k are mandatory." << endl;
} // printUsageInfo() ...

